
## DESCRIPTION ##
#######################################################################
# Copyright (c) 2019 Ian Taylor, Becton Dickinson and Company
#
# Script based on package developed by rezakj, iCellR: https://github.com/rezakj/iCellR
# Plugin for SeqGeq and FlowJo by: Ian Taylor, Josh Luthy, Dylan Hinson, Shahid Siddiq
# of BD Life Sciences, Ashland, OR
#
# License
# The software is distributed under the terms of the
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
#
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability, maintenance and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this
# software.
#######################################################################

## INSTALL REQUIRED PACKAGES ##
#######################
# Install from CRAN

# These are the dependancies to use!!:
# install.packages("iCellR")
# (Required for CellCycle!:)
# install.packages("phateR")
# install.packages("Matrix")

## TRY TO AUTO INSTALL ##

# Install function for packages
packages<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="http://cran.r-project.org")
    require(x,character.only=TRUE)
  }
}

# the Real Deal (aka required packages) #
packages(backports)
packages(iCellR)
packages(Matrix)
#######################

## LOAD INFO FROM FJ/SG ##
#######################

darkness <- DARK_THEME

EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT

parNames <- c(FJ_PARAMS_LIST)
fjOuts <- "FJ_OUTS"
fjDataFilePath <- "FJ_DATA_FILE_PATH"
fjCsvOutputFile <- "FJ_CSV_OUPUT_FILE"
runID <- "FJ_RUNID"
fjDerParName <- "FJ_PAR_DER_PAR_NAME"
# fjK <- FJ_PAR_K_PAR
isSeqGeq <- FJ_ISSEQGEQ

parX <- "PARX"
parY <- "PARY"
parZ <- "PARZ"

catPar01 <- "FJ_CAT01"
catPar01 <- gsub("(\\[)","",gsub("(\\])","",catPar01))
catPar02 <- "FJ_CAT02"
catPar02 <- gsub("(\\[)","",gsub("(\\])","",catPar02))
catPar03 <- "FJ_CAT03"
catPar03 <- gsub("(\\[)","",gsub("(\\])","",catPar03))

pX <- "PARX"
pX <- gsub("(\\[)","",gsub("(\\])","",pX))
pY <- "PARY"
pY <- gsub("(\\[)","",gsub("(\\])","",pY))
pZ <- "PARZ"
pZ <- gsub("(\\[)","",gsub("(\\])","",pZ))

if (catPar01 == "") {
  colorBy = "monochrome"
  pLabel <- paste0(pX," by ",pY," by ",pZ)
  pLabel <- gsub(" ", ".", pLabel)
}else{
  colorBy = "clusters"
  pLabel <- paste0(pX," by ",pY," by ",pZ," colored by ",catPar01)
  pLabel <- gsub(" ", ".", pLabel)
}

## TODO - Deal with duplicate selections...

if ((catPar02 == catPar03) || (catPar02 == catPar01) || (catPar02 == pX) || (catPar02 == pY) || (catPar02 == pZ)){
catPar02 = ""
}

if ((catPar02 != catPar03) && (catPar03 == catPar01 || catPar03 == pX || catPar03 == pY || catPar03 == pZ)){
catPar03 = ""
}

if ((catPar01 == pX) || (catPar01 == pY) || (catPar01 == pZ)){
catPar01 = ""
}


## TODO UNCOMMENT THIS OUT
## set your working directory
 setwd(fjOuts)
#######################

## FUNCTIONS ##
#######################
## For Error Messaging
stopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(fjCsvOutputFile, "_err.txt"))
  stop(message, ...)
}

## LOAD DATA ##
#######################
library(Matrix)
library(iCellR)

  ## TODO UNCOMMENT THIS OUT
   sample <- fjDataFilePath

  ## TODO COMMENT THIS OUT
  ## FOR TESTING ONLY ##
  # setwd("/Users/10201397/Desktop/TODO/icellr_testing")
  # sample <- "/Users/10201397/Desktop/TODO/icellr_testing/exportModTest.csv"

    if (isSeqGeq){
          prePreData <- read.table(file= sample, sep = ",", row.names = 1, header = TRUE, check.names = FALSE)
        } else {
          data <- read.csv(fjDataFilePath, check.names=FALSE)
          cellIdsColumn <- rep(NA, nrow(data))
          cellIds <- data.frame(CellIdDP = cellIdsColumn)
          names(cellIds) <- CellIdDP
          if (CellIdDP %in% names(data)) {
            cellIds <- data[, CellIdDP]
            data <- data[, -which(names(data) %in% c(CellIdDP))]
          }
           EventNumberColumn <- data$EventNumberDP
           eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
           # EventNumberColumn <- rep(NA, nrow(data))
           rownames(data) <- data$EventNumberDP
           prePreData <- data
           colnames(prePreData) <- c(parNames, EventNumberDP)
        }

    if (catPar02 == ""){
          writeLines("No secondary categorical parameter's selected for annotation")
        }else{
          nonUnaryCat <- which(colnames(prePreData) %in% c(catPar02))
          daCat <- subset(prePreData, select = nonUnaryCat)
          daCat <- data.frame(daCat[,1])
          prePreData <- subset(prePreData, select = -nonUnaryCat )

          colnames(daCat) <- catPar02
          colnames(daCat) <- gsub(" ", "", colnames(daCat))

          ## Calibrate Annotation Parameter
          minim <- min(daCat)
          maxim <- max(daCat)
          num.cluste <- nrow(unique(daCat))
          factor <- NULL
          if (min(daCat) == 0){
            factor <- (num.cluste-1)/maxim
          }else{
            factor <- num.cluste/maxim
          }
          calibratedCat <- round(daCat*factor)
          daCat <- calibratedCat
        }


       if (catPar01 == ""){
         writeLines("No secondary categorical parameter's selected for annotation")
       }else{
         nonUnaryCat00 <- which(colnames(prePreData) %in% c(catPar01))
         mainCat <- subset(prePreData, select = nonUnaryCat00)
         mainCat <- data.frame(mainCat[,1])

         prePreData <- subset(prePreData, select = -nonUnaryCat00 )

         colnames(mainCat) <- catPar01
         colnames(mainCat) <- gsub(" ", "", colnames(mainCat))

         ## Calibrate Coloring Parameter
         mini <- min(mainCat)
         maxi <- max(mainCat)
         num.clusti <- nrow(unique(mainCat))
         factor <- NULL
         if (min(mainCat) == 0){
           factor <- (num.clusti-1)/maxi
         }else{
           factor <- num.clusti/maxi
         }
         calibratedColCat <- round(mainCat*factor)
         mainCat <- calibratedColCat
       }

       if (catPar03 == ""){
            writeLines("No cell sizing parameter selected for annotation")
          } else {
            # nonUnaryCat00 <- which(colnames(prePreData) %in% c(catPar01))
            catSize <- subset(prePreData, select = catPar03)
            catSize <- data.frame(catSize[,1])

            # prePreData <- subset(prePreData, select = -nonUnaryCat00 )

            colnames(catSize) <- catPar03
            colnames(catSize) <- gsub(" ", "", colnames(catSize))
          }


    featureAnnotationsFrame <- data.frame(colnames(prePreData))
    cellAnnotationsFrame <- data.frame(rownames(prePreData))

  preData <- t(as.matrix(prePreData))

  t.data <- t(preData)

  if (!isSeqGeq) {
  colnames(t.data) <- gsub(".*? :: (.+)", "\\1",colnames(t.data))
  }

  colnames(t.data) <- gsub(" ", "", colnames(t.data))

  colnames(t.data) <- paste(colnames(t.data),"  Gene Expression")

  data <- t(t.data)


    if (catPar02 == ""){
        colnames(data) <- paste0("CellNamed_", colnames(data))
      }else{
        cellAnnoMod <- cellAnnotationsFrame
        concatMe <- cbind(daCat, cellAnnoMod)
        renameMe <- paste0(catPar02, "=", concatMe[,1], "_CellNamed.", concatMe[,2])
        colnames(data) <- renameMe
      }

  sparse.data <- Matrix(data , sparse = T )
  ##cellIDs look good here...

  dir.create("sgdm")

  dear.dir <- paste0(getwd(),"/sgdm")

  writeMM(obj = sparse.data, file=paste0(getwd(),"/sgdm/matrix.mtx"))
  write.table(t(data[0,]), file=paste0(getwd(),"/sgdm/barcodes.tsv"), col.names=F, row.names=T, quote = F)
  write.table(data[,0], file=paste0(getwd(),"/sgdm/features.tsv"), col.names=F, row.names=T, quote = F)

  data.obj <- load10x(dear.dir,1)
  my.obj <- make.obj(data.obj)
#######################

  ## GET main.data BUILT ##
  #######################

    min.mito = 0
    max.mito = 1
    min.genes = 0
    max.genes = Inf
    min.umis = 0
    max.umis = Inf
    filter.by.cell.id = "character"
    keep.cell.id = "character"
    filter.by.gene = "character"
    filter.by.gene.exp.min = 1

    if ("iCellR" != class(my.obj)[1]) {
      stop("Your Object should be an object of class iCellR")
    }
    # take input from
    DATA <- my.obj@raw.data

    # filter by gene
    if (filter.by.gene[1] != "character"){
      genExist <- dim(subset(DATA, row.names(DATA) %in% filter.by.gene))[1]
      numberOfgenes <- length(filter.by.gene)
      if (genExist != numberOfgenes) {
        stop("your data lacks the gene/genes in filter.by.gene.")
      }
      gene.filt <- t(subset(DATA, row.names(DATA) %in% filter.by.gene) >= filter.by.gene.exp.min)
      # gene.filted <- as.data.frame(rowSums(gene.filt) > 0L)
      gene.filted <- as.data.frame(apply(gene.filt, 1, any))
      colnames(gene.filted) <- "gene.filter"
      gene.filt.cells <- row.names(subset(gene.filted, gene.filter == F))
      goneCells <- length(gene.filt.cells)
      allCells <- length(colnames(DATA))
    }

    GenesForFilter = paste(filter.by.gene, collapse=",")
    if (filter.by.gene[1] == "character") {
      GenesForFilter = "no cells were filtered based on expression."
    }


    ######## filter out ids
    if (filter.by.cell.id[1] != "character"){
      if (keep.cell.id[1] != "character"){
        stop("if filter.by.cell.id is set you can't set keep.cell.id")
      }
      z = filter.by.cell.id %in% colnames(DATA)
      Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
      Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
      if (length(filter.by.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
      }
      DATA <- DATA[ , -which(names(DATA) %in% filter.by.cell.id)]
      FiltCellIds <- paste(filter.by.cell.id , collapse=",")
      message(paste("The following cells were filtered out: ", FiltCellIds))
    }
    if (filter.by.cell.id[1] == "character") {
      FiltCellIds = "no cell ids were filtered."
    }


    ######## keep ids
    if (keep.cell.id[1] != "character"){
      if (filter.by.cell.id[1] != "character"){
        stop("if keep.cell.id is set you can't set filter.by.cell.id")
      }
      z = keep.cell.id %in% colnames(DATA)
      Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
      Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
      if (length(keep.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
      }
      DATA <- DATA[ , which(names(DATA) %in% keep.cell.id)]
      FiltCellIds <- paste(keep.cell.id, collapse=",")
      message(paste("The following cells are kept: ", FiltCellIds))
    }
    if (keep.cell.id[1] == "character") {
      FiltCellIds = "no cell ids to keep."
    }


    # make a filter parameters file
    FilterFile <- paste("Filters are:
- min mito",min.mito,"max mito",max.mito,"
- min # of genes",min.genes,"max # of genes",max.genes,"
- min UMIs",min.umis,"max UMIs",max.umis,"
- Cells were also filtered by the expression of the following genes (user input):
",GenesForFilter,"
- In addition the following cell ids were set to keep/filter (user input):
",FiltCellIds)
    row.names(DATA) <- gsub("-",".", row.names(DATA))
    attributes(my.obj)$main.data <- DATA
    attributes(my.obj)$my.filters <- FilterFile
#######################


  ## PCA ##
  #######################

  data.type = "pca"
  method = "base.mean.rank"
  top.rank = 500
  plus.log.value = 0.1
  batch.norm = FALSE
  gene.list = "character"

    counts.pca <- my.obj@raw.data

    attributes(my.obj@raw.data)$pca.info <- counts.pca
    dataPCA = counts.pca # [1:max.dim]
    attributes(my.obj@main.data)$pca.data <- dataPCA
    attributes(my.obj)$pca.data <- dataPCA
    row.names(dataPCA) <- gsub("-",".", row.names(dataPCA))
    attributes(my.obj)$pca.data <- dataPCA
    dims <- length(prePreData)

#######################


## PLOT IT ##
############
### 3D Plot ###

## Utilize the Dark-Theme (or not):
if (darkness) {

color = "white"
colorb = "black"

} else {

color = "black"
colorb = "white"

}

      x <- my.obj # yoObj
      cell.size = 1

      ## Note: Using PCA as a proxy for the plot type we really want (which is self inflicted, not specifically PCA, tSNE or UMAP)
      plot.type = "pca"
      cell.color = color
      back.col = colorb
      col.by = colorBy # type: "clusters" or "monochrome"
      cond.shape = FALSE
      cell.transparency = 0.5
      clust.dim = 3
      angle = 20
      clonotype.max = 10
      density = FALSE
      interactive = TRUE
      static3D = FALSE
      out.name = plot.type
      z.axis = my.obj@pca.data[,3] # my.obj@pca.data[,3], or my.obj@tsne.data[,1]
      z.axis = t(z.axis)

      f1 <- list(
        family = "Verdana",
        size = 14,
        color = color
      )

      f2 <- list(
        family = "Verdana",
        size = 8,
        color = color
      )

      xie <- list(
        showticklabels = TRUE,
        titlefont = f1,
        tickfont = f2,
        # zeroline = TRUE,
        showline = TRUE,
        # mirror = "ticks",
        gridcolor = color,
        gridwidth = 1,
        # zerolinecolor = toRGB("red"),
        # zerolinewidth = 4,
        # linecolor = toRGB(colorb),
        # linewidth = 1,
        title = pX
      )
      yie <- list(
        showticklabels = TRUE,
        titlefont = f1,
        tickfont = f2,
        # zeroline = TRUE,
        showline = TRUE,
        # mirror = "ticks",
        gridcolor = color,
        gridwidth = 1,
        # zerolinecolor = toRGB("red"),
        # zerolinewidth = 4,
        # linecolor = toRGB(colorb),
        # linewidth = 1,
        title = pY
      )
      zie <- list(
        showticklabels = TRUE,
        titlefont = f1,
        tickfont = f2,
        # zeroline = TRUE,
        showline = TRUE,
        # mirror = "ticks",
        gridcolor = color,
        gridwidth = 1,
        # zerolinecolor = toRGB("red"),
        # zerolinewidth = 4,
        # linecolor = toRGB(colorb),
        # linewidth = 2,
        title = pZ
      )

      ## Note: Using PCA as a proxy for the plot type we really want (which is self inflicted, not specifically PCA, tSNE or UMAP)
            if (plot.type == "pca") {
              MyTitle = pLabel
              DATA <- t(x@pca.data)
              DATA <- as.data.frame(DATA)
            }

if (catPar03 == "") {
            if (col.by == "clusters") {
              library(RColorBrewer)
              col.legend <- (mainCat)
              qual_col_pals = brewer.pal.info[brewer.pal.info$category == 'qual',]
              colors = unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))
              colors <- colors[5:30]

              cols <- colors[as.factor(mainCat[,1])]
                            meta.col.legend <- as.factor(mainCat[,1])
                            col.legend <- paste("Cluster ", meta.col.legend)

              # plot 3d

              # DATAann <- as.data.frame(mainCat$Cluster)
              DATAann <- as.data.frame(mainCat)
              # DATAann <- DATAann
              A = (row.names(DATAann))
              colnames(DATAann) <- ("clusters")
              B = as.character(as.matrix(DATAann[1]))
              ANNOT <- paste(A,"clust",B,sep="_")
              x = DATA[,1]
              myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = DATA[,3],
                                text = row.names(DATA),
                                type="scatter3d", mode = "markers",
                                # titlefont = f1,
                                # tickfont = f2,
                                opacity = cell.transparency, marker = list(size = cell.size + 3), color = col.legend) %>%
                layout(plot_bgcolor = back.col) %>%
                layout(paper_bgcolor = back.col) %>%
                layout(title = MyTitle,
                       scene = list(xaxis = xie,
                                    yaxis = yie,
                                    zaxis = zie)
                )
            }else{ myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = DATA[,3],
                                    text = row.names(DATA),type="scatter3d", mode = "markers",
                                    opacity = cell.transparency,
                                    marker = list(size = cell.size + 3, color = cell.color)) %>%
              layout(plot_bgcolor = back.col) %>%
              layout(paper_bgcolor = back.col) %>%
              layout(title = MyTitle,
                     scene = list(xaxis = xie,
                                  yaxis = yie,
                                  zaxis = zie)
              )
            }
            ## !!! 'Else' we have a quaternary parameter to influence the siz of dots!!! ##
         } else {

         ## TODO - Figure out how to reference that dimension exactly.
         dsize = as.matrix(catSize)
         if (col.by == "clusters") {
                       library(RColorBrewer)
                       col.legend <- (mainCat)
                       qual_col_pals = brewer.pal.info[brewer.pal.info$category == 'qual',]
                       colors = unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))
                       colors <- colors[5:30]

                       cols <- colors[as.factor(mainCat[,1])]
                                     meta.col.legend <- as.factor(mainCat[,1])
                                     col.legend <- paste("Cluster ", meta.col.legend)

                       # plot 3d

                       # DATAann <- as.data.frame(mainCat$Cluster)
                       DATAann <- as.data.frame(mainCat)
                       # DATAann <- DATAann
                       A = (row.names(DATAann))
                       colnames(DATAann) <- ("clusters")
                       B = as.character(as.matrix(DATAann[1]))
                       ANNOT <- paste(A,"clust",B,sep="_")
                       x = DATA[,1]
                       myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = DATA[,3],
                                         text = row.names(DATA),
                                         size = dsize,
                                         type="scatter3d",
                                         mode = "markers",
                                         marker = list(symbol = 'circle',
                                                       sizemode = 'diameter'),
                                         sizes = c(5, 40),
                                         # titlefont = f1,
                                         # tickfont = f2,
                                         opacity = cell.transparency, marker = list(size = cell.size + 3), color = col.legend) %>%
                         layout(plot_bgcolor = back.col) %>%
                         layout(paper_bgcolor = back.col) %>%
                         layout(title = MyTitle,
                                scene = list(xaxis = xie,
                                             yaxis = yie,
                                             zaxis = zie)
                         )
                     }else{ myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = DATA[,3],
                                             text = row.names(DATA),
                                             size = dsize,
                                             type="scatter3d",
                                             mode = "markers",
                                             marker = list(symbol = 'circle',
                                                           sizemode = 'diameter'),
                                             sizes = c(5, 40),
                                             opacity = cell.transparency,
                                             marker = list(size = cell.size + 3, color = cell.color)) %>%
                       layout(plot_bgcolor = back.col) %>%
                       layout(paper_bgcolor = back.col) %>%
                       layout(title = MyTitle,
                              scene = list(xaxis = xie,
                                           yaxis = yie,
                                           zaxis = zie)
                       )
                     }
         }
            OUT.PUT <- paste("3D_FlowJo_Run-", runID, ".html", sep="")
            ## TODO - Get selfcontained = TRUE flag to work in R...
            suppressWarnings(htmlwidgets::saveWidget(ggplotly(myPLOT), OUT.PUT, selfcontained = FALSE, libdir = NULL, title = class(runID)[[1]]))


#################
### SAVE GAME ###
save(my.obj, file = "my.obj.Robj")
#################

write.table(data.frame(1,2,3,5,8), file="fin.tsv", sep="  ", col.names=F, row.names = F)