
## DESCRIPTION ##
#######################################################################
# Copyright (c) 2019 Ian Taylor, Becton Dickinson and Company
#
# Script based on package developed by rezakj, iCellR: https://github.com/rezakj/iCellR
# Plugin for SeqGeq and FlowJo by: Ian Taylor, Josh Luthy, Dylan Hinson, Shahid Siddiq,
# Matt Swindle, and Richard Halpert of BD Life Sciences, Ashland, OR
#
# License
# The software is distributed under the terms of the
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
#
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability, maintenance and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this
# software.
#######################################################################

## INSTALL REQUIRED PACKAGES ##
#######################
# Install from CRAN

# These are the dependancies to use!!:
# install.packages("iCellR")
# (Required for CellCycle!:)
# install.packages("phateR")
# install.packages("Matrix")

## TRY TO AUTO INSTALL ##

# Install function for packages
packages<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="http://cran.r-project.org")
    require(x,character.only=TRUE)
  }
}

# the Real Deal (aka required packages) #
packages(backports)
packages(iCellR)
packages(Matrix)
#######################

## LOAD INFO FROM FJ/SG ##
#######################

EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT

parNames <- c(FJ_PARAMS_LIST)
fjOuts <- "FJ_OUTS"
fjDataFilePath <- "FJ_DATA_FILE_PATH"
fjCsvOutputFile <- "FJ_CSV_OUPUT_FILE"
runID <- "FJ_RUNID"
fjDerParName <- "FJ_PAR_DER_PAR_NAME"
# fjK <- FJ_PAR_K_PAR
isSeqGeq <- FJ_ISSEQGEQ
fQVal <- FJ_fQVal
fFC <- FJ_FC
catPar <- "FJ_CAT"
catPar <- gsub("(\\[)","",gsub("(\\])","",catPar))

clustMtx <- "FJ_CLUST"
# "ward.D", "ward.D2", "single", "complete", "average", "mcquitty", "median", "centroid", "kmeans", "NONE"

clustDist <- "FJ_DIST"
  # "euclidean", "maximum", "manhattan", "canberra", "binary", "minkowski" or "NULL"

clustIdx <- "FJ_IDX"
  # "kl", "ch", "hartigan", "ccc", "scott", "marriot", "trcovw", "tracew", "friedman", "rubin",
  # "cindex", "db", "silhouette", "duda", "pseudot2", "beale", "ratkowsky", "ball", "ptbiserial",
  # "gap", "frey", "mcclain", "gamma", "gplus", "tau", "dunn", "hubert", "sdindex", "dindex", "sdbw"


if (clustMtx == "NONE") {
  colorBy = "monochrome"
}else{
  colorBy = "clusters"
}

## TODO UNCOMMENT THIS OUT
## set your working directory
 setwd(fjOuts)
#######################

## FUNCTIONS ##
#######################
## For Error Messaging
stopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(fjCsvOutputFile, "_err.txt"))
  stop(message, ...)
}

makeYo3DPlot <- function(yoObj, type, zAxis, slfCntd) { # colorBy - Not needed any longer; slfCntd is a boolean "TRUE" or "FALSE"
  x <- yoObj # yoObj
  cell.size = 1
  plot.type = type # plot: "pca", "tsne", or "umap"
  cell.color = "black"
  back.col = "white"
  col.by = colorBy # type: "clusters" or "monochrome"
  cond.shape = FALSE
  cell.transparency = 0.5
  clust.dim = 3
  angle = 20
  clonotype.max = 10
  density = FALSE
  interactive = TRUE
  static3D = FALSE
  out.name = plot.type
  z.axis = zAxis # my.obj@pca.data[,3], or my.obj@tsne.data[,1]


  if (plot.type == "tsne") {
    MyTitle = "tSNE"
    DATA <- x@tsne.data.3d
  }
  if (plot.type == "pca") {
    MyTitle = "PCA"
    DATA <- x@pca.data
  }
  if (plot.type == "umap") {
    MyTitle = "UMAP"
    DATA <- x@umap.data
  }
  if (plot.type == "diffusion") {
    MyTitle = "Diffusion Map"
    DATA <- x@diffusion.data
  }

  if (col.by == "clusters") {
    library(RColorBrewer)
    col.legend <- factor(x@best.clust$clusters)
    qual_col_pals = brewer.pal.info[brewer.pal.info$category == 'qual',]
    colors = unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))
    colors <- colors[5:30]
    cols <- colors[as.numeric(x@best.clust$clusters)]
    col.legend <- paste0("Cluster ", as.factor(x@best.clust$clusters))

    # plot 3d
    DATAann <- as.data.frame(x@cluster.data$Best.partition)
    DATAann <- as.data.frame(x@best.clust)
    A = (row.names(DATAann))
    colnames(DATAann) <- ("clusters")
    B = as.character(as.matrix(DATAann[1]))
    ANNOT <- paste(A,"clust",B,sep="_")
    myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = z.axis, #text = ANNOT,#
                      text = row.names(DATAann),
                      type="scatter3d",
                      mode = "markers",
                      opacity = cell.transparency,
                      marker = list(size = cell.size + 4),
                      color = col.legend) %>%
      layout(plot_bgcolor = back.col) %>%
      layout(paper_bgcolor = back.col) %>%
      layout(title = MyTitle,
             scene = list(xaxis = list(title = paste0( MyTitle , "_X" )),
                          yaxis = list(title = paste0( MyTitle , "_Y" )),
                          zaxis = list(title = paste0( MyTitle , "_Z" )))
      )
  }else{myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = z.axis,
                          text = row.names(DATAann),type="scatter3d", mode = "markers", opacity = cell.transparency,
                          marker = list(size = cell.size + 2, color = cell.color)) %>%
    layout(DATA, x = DATA[,1], y = DATA[,2], z = z.axis) %>%
    layout(plot_bgcolor = back.col) %>%
    layout(paper_bgcolor = back.col) %>%
    layout(title = MyTitle,
           scene = list(xaxis = list(title = "Dim1"),
                        yaxis = list(title = "Dim2"),
                        zaxis = list(title = "Dim3"))
    )
  }
  OUT.PUT <- paste(out.name, ".html", sep="")
  suppressWarnings(htmlwidgets::saveWidget(ggplotly(myPLOT), OUT.PUT, selfcontained = slfCntd, libdir = NULL, title = class(runID)[[1]]))
}

## TODO - Make size a thing you can adjust!
makeYo3DPlotSz <- function(yoObj, type, zAxis, slfCntd, dsize) { # colorBy - Not needed any longer; slfCntd is a boolean "TRUE" or "FALSE"
  x <- yoObj # yoObj
  cell.size = 1
  plot.type = type # plot: "pca", "tsne", or "umap"
  cell.color = "black"
  back.col = "white"
  col.by = colorBy # type: "clusters" or "monochrome"
  cond.shape = FALSE
  cell.transparency = 0.5
  clust.dim = 3
  angle = 20
  clonotype.max = 10
  density = FALSE
  interactive = TRUE
  static3D = FALSE
  out.name = plot.type
  z.axis = zAxis # my.obj@pca.data[,3], or my.obj@tsne.data[,1]
  dsize <- as.matrix(dsize)

      if (plot.type == "tsne") {
        MyTitle = "tSNE"
        DATA <- x@tsne.data.3d
      }
      if (plot.type == "pca") {
        MyTitle = "PCA"
        DATA <- x@pca.data
      }
      if (plot.type == "umap") {
        MyTitle = "UMAP"
        DATA <- x@umap.data
      }
      if (plot.type == "diffusion") {
        MyTitle = "Diffusion Map"
        DATA <- x@diffusion.data
      }

    if (dsize != "")
      if (col.by == "clusters") {
        library(plotly)
        library(RColorBrewer)
        col.legend <- factor(x@best.clust$clusters)
        qual_col_pals = brewer.pal.info[brewer.pal.info$category == 'qual',]
        colors = unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))
        colors <- colors[5:30]
        cols <- colors[as.numeric(x@best.clust$clusters)]
        col.legend <- paste0("Cluster ", as.factor(x@best.clust$clusters))
        DATAann <- as.data.frame(x@cluster.data$Best.partition)
        DATAann <- as.data.frame(x@best.clust)
        A = (row.names(DATAann))
        colnames(DATAann) <- ("clusters")
        B = as.character(as.matrix(DATAann[1]))
        ANNOT <- paste(A,"clust",B,sep="_")
        myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = z.axis, #text = ANNOT,#
                          text = row.names(DATAann),
                          size = dsize,
                          type="scatter3d",
                          mode = "markers",
                          opacity = cell.transparency,
                          marker = list(symbol = 'circle',
                                        sizemode = 'diameter'),
                          sizes = c(5, 90),
                          color = col.legend) %>%
          layout(plot_bgcolor = back.col) %>%
          layout(paper_bgcolor = back.col) %>%
          layout(title = MyTitle,
                 scene = list(xaxis = list(title = paste0( MyTitle , "_X" )),
                              yaxis = list(title = paste0( MyTitle , "_Y" )),
                              zaxis = list(title = paste0( MyTitle , "_Z" )))
          )
      }else{myPLOT <- plot_ly(DATA, x = DATA[,1], y = DATA[,2], z = z.axis,
                              text = row.names(DATAann),
                              size = dsize,
                              type="scatter3d",
                              mode = "markers",
                              opacity = cell.transparency,
                              marker = list(symbol = 'circle',
                                            sizemode = 'diameter'),
                              sizes = c(5, 90),
                              color = cell.color) %>%
        layout(DATA, x = DATA[,1], y = DATA[,2], z = z.axis) %>%
        layout(plot_bgcolor = back.col) %>%
        layout(paper_bgcolor = back.col) %>%
        layout(title = MyTitle,
               scene = list(xaxis = list(title = "Dim1"),
                            yaxis = list(title = "Dim2"),
                            zaxis = list(title = "Dim3"))
        )
      }
      OUT.PUT <- paste(out.name, ".html", sep="")
      suppressWarnings(htmlwidgets::saveWidget(ggplotly(myPLOT), OUT.PUT, selfcontained = slfCntd, libdir = NULL, title = class(runID)[[1]]))
    }




## LOAD DATA ##
#######################
library(Matrix)
library(iCellR)

  ## TODO UNCOMMENT THIS OUT
   sample <- fjDataFilePath

  ## TODO COMMENT THIS OUT
  ## FOR TESTING ONLY ##
  # setwd("/Users/10201397/Desktop/TODO/icellr_testing")
  # sample <- "/Users/10201397/Desktop/TODO/icellr_testing/exportModTest.csv"

    if (isSeqGeq){
          prePreData <- read.table(file= sample, sep = ",", row.names = 1, header = TRUE, check.names = FALSE)
        } else {
          data <- read.csv(fjDataFilePath, check.names=FALSE)
          cellIdsColumn <- rep(NA, nrow(data))
          cellIds <- data.frame(CellIdDP = cellIdsColumn)
          names(cellIds) <- CellIdDP
          if (CellIdDP %in% names(data)) {
            cellIds <- data[, CellIdDP]
            data <- data[, -which(names(data) %in% c(CellIdDP))]
          }
           EventNumberColumn <- data$EventNumberDP
           eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
           # EventNumberColumn <- rep(NA, nrow(data))
           rownames(data) <- data$EventNumberDP
           prePreData <- data
           colnames(prePreData) <- c(parNames, EventNumberDP)
        }

        if (catPar == "") {
                  writeLines("No categorical parameter's selected for annotation")
                } else {
                  if (isSeqGeq) {
                  nonUnaryCat <- which(colnames(prePreData) %in% c(catPar))
                  daCat <- subset(prePreData, select = nonUnaryCat)
                  daCat <- data.frame(daCat[,1])
                  prePreData <- subset(prePreData, select = -nonUnaryCat )
                  colnames(daCat) <- gsub(" ", "", colnames(daCat))
                  } else {
                    nonUnaryCat <- which(colnames(prePreData) %in% c(catPar))
                    daCat <- subset(prePreData, select = nonUnaryCat)
                    daCat <- data.frame(daCat[,1])
                    prePreData <- subset(prePreData, select = -nonUnaryCat )
                    colnames(daCat) <- gsub(" ", "", colnames(daCat))
                  }

                }

    featureAnnotationsFrame <- data.frame(colnames(prePreData))
    cellAnnotationsFrame <- data.frame(rownames(prePreData))
  
  preData <- t(as.matrix(prePreData))
  
  t.data <- t(preData)

  if (!isSeqGeq) {
  colnames(t.data) <- gsub(".*? :: (.+)", "\\1",colnames(t.data))
  }

  colnames(t.data) <- gsub(" ", "", colnames(t.data))

  colnames(t.data) <- paste(colnames(t.data),"  Gene Expression")
  
  data <- t(t.data)

    if (catPar == ""){
        colnames(data) <- paste0("CellNamed_", colnames(data))
      }else{


     ## Calibrate Annotation Parameter
         minim <- min(daCat)
         maxim <- max(daCat)
         num.cluste <- nrow(unique(daCat))
         factor <- NULL
         if (min(daCat) == 0){
           factor <- (num.cluste-1)/maxim
         }else{
           factor <- num.cluste/maxim
         }
         calibratedCat <- round(daCat*factor)
         daCat <- calibratedCat

     ## Add Annotations to CellIDs
        cellAnnoMod <- cellAnnotationsFrame
        concatMe <- cbind(daCat, cellAnnoMod)
        renameMe <- paste0(catPar, "=", concatMe[,1], "_CellNamed.", concatMe[,2])
        colnames(data) <- renameMe
      }

  sparse.data <- Matrix(data , sparse = T )
  ##cellIDs look good here...
  
  dir.create("sgdm") 
  
  dear.dir <- paste0(getwd(),"/sgdm")
  
  writeMM(obj = sparse.data, file=paste0(getwd(),"/sgdm/matrix.mtx"))
  write.table(t(data[0,]), file=paste0(getwd(),"/sgdm/barcodes.tsv"), col.names=F, row.names=T, quote = F)
  write.table(data[,0], file=paste0(getwd(),"/sgdm/features.tsv"), col.names=F, row.names=T, quote = F)
  
  data.obj <- load10x(dear.dir,1)
  my.obj <- make.obj(data.obj)
#######################
  
  ## GET main.data BUILT ##
  #######################
    
    min.mito = 0
    max.mito = 1
    min.genes = 0
    max.genes = Inf
    min.umis = 0
    max.umis = Inf
    filter.by.cell.id = "character"
    keep.cell.id = "character"
    filter.by.gene = "character"
    filter.by.gene.exp.min = 1
    
    if ("iCellR" != class(my.obj)[1]) {
      stop("Your Object should be an object of class iCellR")
    }
    # take input from
    DATA <- my.obj@raw.data
    
    # filter by gene
    if (filter.by.gene[1] != "character"){
      genExist <- dim(subset(DATA, row.names(DATA) %in% filter.by.gene))[1]
      numberOfgenes <- length(filter.by.gene)
      if (genExist != numberOfgenes) {
        stop("your data lacks the gene/genes in filter.by.gene.")
      }
      gene.filt <- t(subset(DATA, row.names(DATA) %in% filter.by.gene) >= filter.by.gene.exp.min)
      # gene.filted <- as.data.frame(rowSums(gene.filt) > 0L)
      gene.filted <- as.data.frame(apply(gene.filt, 1, any))
      colnames(gene.filted) <- "gene.filter"
      gene.filt.cells <- row.names(subset(gene.filted, gene.filter == F))
      goneCells <- length(gene.filt.cells)
      allCells <- length(colnames(DATA))
    }
    #
    GenesForFilter = paste(filter.by.gene, collapse=",")
    if (filter.by.gene[1] == "character") {
      GenesForFilter = "no cells were filtered based on expression."
    }
    # filter by cell id
    # filter.by.cell.id = c("WT_AAACATACAACCAC.1","WT_AAACATTGATCAGC.1")
    ######## filter out ids
    if (filter.by.cell.id[1] != "character"){
      if (keep.cell.id[1] != "character"){
        stop("if filter.by.cell.id is set you can't set keep.cell.id")
      }
      z = filter.by.cell.id %in% colnames(DATA)
      Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
      Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
      if (length(filter.by.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
      }
      DATA <- DATA[ , -which(names(DATA) %in% filter.by.cell.id)]
      FiltCellIds <- paste(filter.by.cell.id , collapse=",")
      message(paste("The following cells were filtered out: ", FiltCellIds))
    }
    if (filter.by.cell.id[1] == "character") {
      FiltCellIds = "no cell ids were filtered."
    }
    ######## keep ids
    if (keep.cell.id[1] != "character"){
      if (filter.by.cell.id[1] != "character"){
        stop("if keep.cell.id is set you can't set filter.by.cell.id")
      }
      z = keep.cell.id %in% colnames(DATA)
      Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
      Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
      if (length(keep.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
      }
      DATA <- DATA[ , which(names(DATA) %in% keep.cell.id)]
      FiltCellIds <- paste(keep.cell.id, collapse=",")
      message(paste("The following cells are kept: ", FiltCellIds))
    }
    if (keep.cell.id[1] == "character") {
      FiltCellIds = "no cell ids to keep."
    }
    
    
    # make a filter parameters file
    FilterFile <- paste("Filters are:
- min mito",min.mito,"max mito",max.mito,"
- min # of genes",min.genes,"max # of genes",max.genes,"
- min UMIs",min.umis,"max UMIs",max.umis,"
- Cells were also filtered by the expression of the following genes (user input):
",GenesForFilter,"
- In addition the following cell ids were set to keep/filter (user input):
",FiltCellIds)
    # write filter parameters
    # write.table((FilterFile),file="filters_set.txt", row.names =F, quote = F, col.names = F)
    # message("filters_set.txt file has beed generated and includes the filters set for this experiment.")
    # return data
    row.names(DATA) <- gsub("-",".", row.names(DATA))
    attributes(my.obj)$main.data <- DATA
    attributes(my.obj)$my.filters <- FilterFile
#######################


  ## PCA ##
  #######################

  data.type = "pca"
  method = "base.mean.rank"
  top.rank = 500
  plus.log.value = 0.1
  batch.norm = FALSE
  gene.list = "character"
  
    raw.data.order <- my.obj@raw.data[ order(rowMeans(my.obj@raw.data), decreasing = TRUE), ]

    if (isSeqGeq){
        topGenes <- (raw.data.order)
        TopNormLogScale <- log(topGenes + plus.log.value)
        } else {
        topGenes <- (raw.data.order)
        sup <- min(topGenes)
        dawg <- topGenes + abs(sup)
        TopNormLogScale <- log(dawg + plus.log.value)
        }

    counts.pca <- prcomp(TopNormLogScale, center = FALSE, scale. = FALSE)
    attributes(my.obj@raw.data)$pca.info <- counts.pca
    # DATA
    dataPCA = data.frame(counts.pca$rotation) # [1:max.dim]
    attributes(my.obj@main.data)$pca.data <- dataPCA
    # optimal
    pcaDATA <- counts.pca$sdev
    OPTpcs <- mean(pcaDATA)*2
    OPTpcs <- (pcaDATA > OPTpcs)
    OPTpcs <- length(OPTpcs[OPTpcs==TRUE]) + 1
    attributes(my.obj@main.data)$opt.pcs <- OPTpcs
    attributes(my.obj)$pca.data <- dataPCA
    row.names(dataPCA) <- gsub("-",".", row.names(dataPCA))
    attributes(my.obj)$pca.data <- dataPCA

    if (length(dataPCA) < 10) {
      dims <- length(dataPCA)
    } else {
      dims <- 10
    }

#######################


## CLUSTERING ##
  #######################

  ## NOTES: ##

  # Here we cluster the first 10 dimensions of the data which is converted to principal
  # components. You have the option of clustering your data based on the following methods:
  # "ward.D", "ward.D2", "single", "complete", "average", "mcquitty", "median", "centroid", "kmeans"

  # For the distance calculation used for clustering, you have the following options:
  # "euclidean", "maximum", "manhattan", "canberra", "binary", "minkowski" or "NULL"

  # With the following indexing methods:
  # "kl", "ch", "hartigan", "ccc", "scott", "marriot", "trcovw", "tracew", "friedman", "rubin",
  # "cindex", "db", "silhouette", "duda", "pseudot2", "beale", "ratkowsky", "ball", "ptbiserial",
  # "gap", "frey", "mcclain", "gamma", "gplus", "tau", "dunn", "hubert", "sdindex", "dindex", "sdbw"

## This next step is slow... ##

if(clustMtx != "NONE"){
    clust.method = clustMtx
    dist.method = clustDist
    index.method = clustIdx
    max.clust = 25
    min.clust = 2
    #dims = 1:dims

    library(NbClust)
    DATA <- (my.obj@pca.data)[1:dims]
    nb <- NbClust(DATA,
                distance = dist.method,
                min.nc = min.clust,
                max.nc = max.clust,
                method = clust.method,
                index = index.method)
    cluster.data <- nb
    attributes(my.obj)$cluster.data <- cluster.data
    ####
    clustInfo = as.data.frame(my.obj@cluster.data$Best.partition)
    colnames(clustInfo) <- "clusters"
    attributes(my.obj)$best.clust <- clustInfo
    # my.obj <- clust.avg.exp(my.obj)
    num.clust <- cluster.data[["Best.nc"]][["Number_clusters"]]
    ili <- data.frame(cluster.data$Best.partition)
    il <- sapply(ili, function(x) length(unique(x)) > 1)
    }
  #######################


## DIFFERENTIAL EXPRESSION ##
#############################


if(il == FALSE){
clustMtx = "NONE"
colorBy = "monochrome"
writeLines("No clusters found, try another clustering method")
}

if(clustMtx != "NONE"){

if(isSeqGeq){

  ## TODO - Get other methods of DEG working (ANOVA, etc)
  UqValue <- fQVal
  fcT <- fFC
  fFCinv <- 1/fFC
  fFCinv <- round( fFCinv, 4)

#### Create not gates for each cluster:
for(n in 1:num.clust){
  o <- n
  p <- c()
  for (i in 1:num.clust){
    if (i !=o){
      p <- c(p,i)
    }
  }

  a <- table(my.obj@best.clust$clusters)
  a.num <- a[names(a)==o]

  if (a.num > 1){
##### Calculate DEG for one cluster:
diff.res <- run.diff.exp(my.obj, de.by = "clusters", cond.1 = p, cond.2 = c(n))

##### Write out DEG parameterSets as CSV files:
DEG <- data.frame(featureAnnotationsFrame)
FC <- data.frame(diff.res[,"foldChange"])
PV <- data.frame(diff.res[,"padj"])

DEG <- cbind(DEG, FC, PV)

 qValTH <- subset(DEG, (DEG$diff.res....padj.. < fQVal))
 upFCnQVal <- subset(qValTH, qValTH$diff.res....foldChange.. > fFC)
 downFCnQVal <- subset(qValTH, qValTH$diff.res....foldChange.. < fFCinv)
 nam <- paste("Up_in", clustMtx, n, sep = "_")
 namAlt <- paste("Down_in", clustMtx, n, sep = "_")
 assign(nam, upFCnQVal)
 assign(namAlt, downFCnQVal)

 namGS <- paste0("/Up_in_", clustMtx, "_", n, "_", runID, ".csv")
 fGSnam <- paste0(fjOuts,namGS)
 write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
 write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
 write.table(paste0("$GeneSetName,Up in ", clustMtx, " ", n, " RunID: ", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
 write.table(paste0("$GeneSetDescription,Generated in the iCellR Pipeline. Fold-Change > ",fFC,"; q-Val (FDR) &lt; ", fQVal), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
 write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
 write.table("[Genes],Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
 write.table(upFCnQVal, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")

 namGSd <- paste0("/Down_in_", clustMtx, "_", n, "_", runID, ".csv")
  fGSnamd <- paste0(fjOuts,namGSd)
  write.table("[GeneSet]", file = fGSnamd, append = F, quote = F, row.names = F, col.names = F )
  write.table("$SchemaVersion,1", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
  write.table(paste0("$GeneSetName,Down in ", clustMtx, " ", n, " RunID: ", runID), file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
  write.table(paste0("$GeneSetDescription,Generated in the iCellR Pipeline. Fold-Change &lt; ", fFCinv, "; q-Val (FDR) &lt; ", fQVal), file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
  write.table(" ", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
  write.table("[Genes],Fold-Change,q-Value", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
  write.table(downFCnQVal, file = fGSnamd, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")

}else{
  writeLines(paste0("Cluster number ", n, " contains one or less events."))
}
}
}
}
#######################
#######################


## PLOT IT ##
############
### PCA ###

  # calibrate the parameters so we don't have negative values for scaling later
    ncells <- nrow(cellAnnotationsFrame)
    cal.new.PCA <- data.frame(matrix(nrow=ncells))
    for (i in 1:dims) {
      min.i <- min(my.obj@pca.data[i])
      cali.i <- (my.obj@pca.data[i] - min.i)
      cal.new.PCA[,ncol(cal.new.PCA)+1] <- cali.i
    }
    cal.new.PCA <- cal.new.PCA[,-1]
    attributes(my.obj)$pca.data <- cal.new.PCA

makeYo3DPlot(my.obj, "pca", my.obj@pca.data[,3], FALSE)

############
### TSNE ###
my.obj <- run.pc.tsne(my.obj, dims = 1:dims)

# calibrate the parameters so we don't have negative values for scaling later
ncells <- nrow(cellAnnotationsFrame)
cal.new.tSNE <- data.frame(matrix(nrow=ncells))
for (j in 1:3) {
  min.i <- min(my.obj@tsne.data.3d[j])
  cali.i <- (my.obj@tsne.data.3d[j] - min.i)
  cal.new.tSNE[,ncol(cal.new.tSNE)+1] <- cali.i
}
cal.new.tSNE <- cal.new.tSNE[,-1]
attributes(my.obj)$tsne.data.3d <- cal.new.tSNE


makeYo3DPlot(my.obj, "tsne", my.obj@tsne.data.3d[,3], FALSE)

############
### UMAP ###
my.obj <- run.umap(my.obj, dims = 1:dims) #, method = "naive")

# calibrate the parameters so we don't have negative values for scaling later
ncells <- nrow(cellAnnotationsFrame)
cal.new.UMAP <- data.frame(matrix(nrow=ncells))
for (i in 1:2) {
  min.i <- min(my.obj@umap.data[i])
  cali.i <- (my.obj@umap.data[i] - min.i)
  cal.new.UMAP[,ncol(cal.new.UMAP)+1] <- cali.i
}
cal.new.UMAP <- cal.new.UMAP[,-1]
attributes(my.obj)$umap.data <- cal.new.UMAP

makeYo3DPlot(my.obj, "umap", my.obj@tsne.data[,1], FALSE)

#################
### SAVE GAME ###
save(my.obj, file = "my.obj.Robj")
#################


##### Parameter and Feature Prep
#################################################################

# Create the Derived Parameters Data matrices

q <- nrow(cellAnnotationsFrame)

if(clustMtx != "NONE"){
meta.clust <- data.frame(cluster.data[["Best.partition"]])
write.table(num.clust, paste0("Estimated_k_", runID), row.names = F, col.names = F)

# Data matrix for export to FJ/SG
clust <- cbind(cellAnnotationsFrame, meta.clust[1:q,], my.obj@tsne.data.3d[1:q,], my.obj@umap.data[1:q,], dataPCA[1:q,1:dims])

if (isSeqGeq){
col01 <- c("CellId",paste0(clustMtx, "_", clustDist, "_", clustIdx, "_", runID), paste0("tsne_x_", runID), paste0("tsne_y_", runID), paste0("tsne_z_", runID), paste0("umap_x_", runID), paste0("umap_y_", runID))
}else{
col01 <- c("EventNumberDP",paste0(clustMtx, "_", clustDist, "_", clustIdx, "_", runID), paste0("tsne_x_", runID), paste0("tsne_y_", runID), paste0("tsne_z_", runID), paste0("umap_x_", runID), paste0("umap_y_", runID))
}

} else {

clust <- cbind(cellAnnotationsFrame, my.obj@tsne.data.3d[1:q,], my.obj@umap.data[1:q,], dataPCA[1:q,1:dims])

if (isSeqGeq){
col01 <- c("CellId", paste0("tsne_x_", runID), paste0("tsne_y_", runID), paste0("tsne_z_", runID), paste0("umap_x_", runID), paste0("umap_y_", runID))
}else{
col01 <- c("EventNumberDP", paste0("tsne_x_", runID), paste0("tsne_y_", runID), paste0("tsne_z_", runID), paste0("umap_x_", runID), paste0("umap_y_", runID))
}
}

# Column headers, aka Derived Parameter Names


col_n <- col01
for(d in 1:dims){
  col_n <- c(col_n, paste0("pc_",d,"_",runID))
}
col02 <- col_n
colnames(clust) <- col02

# Create the DP output (CSV)
FName <- paste0("/iCellR_Clust_",runID,".csv")
fileName = paste0(fjOuts,FName)

if (isSeqGeq){
suppressWarnings(write.table(clust, file = fileName, sep = ',', append = T, quote = F, row.names = F, col.names = T ))
}else{
FName <- paste0("/iCellR_Clust_",runID,".R")
fileName = paste0(fjOuts,FName)
suppressWarnings(write.table(clust, file = fileName, sep = ',', append = T, quote = F, row.names = F, col.names = T ))
}

#################################################################
#####


write.table(data.frame(1,2,3,5,8), file="fin.tsv", sep="  ", col.names=F, row.names = F)

