
## DESCRIPTION ##
#######################################################################
# Copyright (c) 2019 Ian Taylor, Becton Dickinson and Company
#
# Script based on package developed by rezakj, iCellR: https://github.com/rezakj/iCellR
# Plugin for SeqGeq and FlowJo by: Ian Taylor, Josh Luthy, Dylan Hinson, Shahid Siddiq
# of BD Life Sciences, Ashland, OR
#
# License
# The software is distributed under the terms of the
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
#
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability, maintenance and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this
# software.
#######################################################################

## INSTALL REQUIRED PACKAGES ##
#######################
# Install from CRAN

# These are the dependancies to use!!:
# install.packages("iCellR")
# (Required for CellCycle!:)
# install.packages("phateR")
# install.packages("Matrix")

## TRY TO AUTO INSTALL ##

# Install function for packages
packages<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="http://cran.r-project.org")
    require(x,character.only=TRUE)
  }
}

# the Real Deal (aka required packages) #
packages(backports)
packages(iCellR)
packages(Matrix)
library(Matrix)
library(iCellR)

#######################

## LOAD INFO FROM FJ/SG ##
#######################

EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT

# parNames <- c(FJ_PARAMS_LIST)
fjOuts <- "FJ_OUTS"
fjDataFilePath <- "FJ_DATA_FILE_PATH"
fjCsvOutputFile <- "FJ_CSV_OUPUT_FILE"
runID <- "FJ_RUNID"
fjDerParName <- "FJ_PAR_DER_PAR_NAME"
fFC <- FJ_FC
fQVal <- FJ_fQVal
isSeqGeq <- FJ_ISSEQGEQ
catPar <- "FJ_CAT"
catParTwo <- "CAT_PAR_TWO"

popMap <- "CATPOP_MAP_ONE"
if(catPar != "{}" && catPar != ""){
    popMap <- gsub(", ", "C=O=D=E=X", popMap)
    popMap <- gsub(" ", "_", popMap)
    popMap <- gsub("C=O=D=E=X", ", ", popMap)
    popMap <- gsub("\\.", "_", popMap)
    # popMap <- gsub("\\=", "Eq", popMap)
}

popMapTwo <- "CATPOP_MAP_TWO"
if(catParTwo != "{}" && catParTwo != ""){
    popMapTwo <- gsub(", ", "C=O=D=E=X", popMapTwo)
    popMapTwo <- gsub(" ", "_", popMapTwo)
    popMapTwo <- gsub("C=O=D=E=X", ", ", popMapTwo)
    popMapTwo <- gsub("\\.", "_", popMapTwo)
    # popMapTwo <- gsub("\\=", "Eq", popMapTwo)
}

catPar <- gsub("(\\[)","",gsub("(\\])","",catPar))
catParRaw <- catPar
catParTwo <- gsub("(\\[)","",gsub("(\\])","",catParTwo))
catParTwoRaw <- catParTwo



## TODO UNCOMMENT THIS OUT
## set your working directory
 setwd(fjOuts)
#######################

## FUNCTIONS ##
#######################
## For Error Messaging
stopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(fjCsvOutputFile, "_err.txt"))
  stop(message, ...)
}

## LOAD DATA ##
#######################

  ## TODO UNCOMMENT THIS OUT
   sample <- fjDataFilePath

  ## TODO COMMENT THIS OUT
  ## FOR TESTING ONLY ##
  # setwd("/Users/10201397/Desktop/TODO/icellr_testing")
  # sample <- "/Users/10201397/Desktop/TODO/icellr_testing/exportModTest.csv"

  prePreData <- read.table(file= sample, sep = ",", row.names = 1, header = TRUE, check.names = FALSE)



if(catParTwo != "{}" && catParTwo != ""){
    categoricalTwo <- prePreData[catParTwo]
}



## TODO - Make population naming happen!!!...
if (catPar != ""){

    ## Remove the "." dots from categorical parameters too...
    catParamTemp <- gsub("\\.","_",catPar)
    catParamTemp <- gsub("\\s","_",catParamTemp)
    catParamTemp <- gsub("\\=", "Eq", catParamTemp)
    names(prePreData)[names(prePreData) == catPar] <- catParamTemp
    catPar <- catParamTemp

    categorical <- prePreData[catPar]

    ## SET UP the POP NAMES
    hasPopMap <- F
    if(popMap != "" && popMap != "{}" && popMap != "[]"){
        hasPopMap <- T

        popMapNew <- gsub("\\}","",gsub("\\{","",gsub("\\s","",popMap)))
        splitMap <- as.data.frame(strsplit(popMapNew,","))
        dataMap <- data.frame( Value = character(),
        Name = character(),
        stringsAsFactors = F )

        for(x in 1:nrow(splitMap)){
            xthSet <- as.data.frame(strsplit(as.character(splitMap[x,]),"="))
            dataMap[x,] <- data.frame(Value = as.character(round(as.numeric(as.character(xthSet[1,])),3) ), ## sig figs set to three to match what java is doing. NOTE: This is all for the sake of FlowSOM
            Name = as.character(xthSet[2,]),
            stringsAsFactors = F)
        }

        categorical <- prePreData[catPar]
        newCategorical <- data.frame("xyz" = character(),
        stringsAsFactors = F)
        colnames(newCategorical) <- catPar

        for (i in 1:nrow(categorical)) {
            flag = 1
            for(j in 1:nrow(dataMap)) {

                if(flag > 0){
                    placeholder = "UNCALLED_CAT"      ## This will allow us to check if the categorical was found
                    catSwap = categorical[i,]         ## If that category isn't found, this will allow us to put something sensible there
                }

                if(categorical[i,] == dataMap[j,1]){
                    placeholder = dataMap[j,2]
                    flag = 0
                }

                if(flag > 0 && placeholder == "UNCALLED_CAT"){              ## Assuming that some round of the 'j' loop finds no categorical match, we will (for a moment) automatically call the category as unknown...
                    newCategorical[i,] = paste0(catPar,"_",catSwap)
                } else if (flag == 0 && placeholder != "UNCALLED_CAT"){     ## ... However, assuming that we eventually find a matching categorical, we will name that cell's population correctly!
                    newCategorical[i,] <- placeholder
                }
            }
        }

        prePreData[catPar] <- newCategorical
    }else {

        catList <- unique(categorical)

        dataMap = data.frame(
        Value = as.numeric(),
        Name = as.character(),
        stringsAsFactors = F
        )

        anInt <- 1
        for(value in catList[[catPar]]){
            dataMap[anInt,] <- data.frame(Value = as.numeric(value),
                                        Name = paste0(catPar,"-", as.character(value)),
                                        stringsAsFactors = F)
            anInt <- anInt + 1
        }


    }
}

if (catParTwo != ""){

    ## Remove the "." dots from categorical parameters too...
    catParamTempTwo <- gsub("\\.","_",catParTwo)
    catParamTempTwo <- gsub("\\s","_",catParamTempTwo)
    catParamTempTwo <- gsub("\\=", "Eq", catParamTempTwo)
    names(prePreData)[names(prePreData) == catParTwo] <- catParamTempTwo
    catParTwo <- catParamTempTwo

    categoricalTwo <- prePreData[catParTwo]

    ## SET UP the second POP NAMES
    hasSecondPopMap <- F
    if(popMapTwo != "" && popMapTwo != "{}" && popMapTwo != "[]"){
        hasSecondPopMap <- T

        popMapNewTwo <- gsub("\\}","",gsub("\\{","",gsub("\\s","",popMapTwo)))
        splitMapTwo <- as.data.frame(strsplit(popMapNewTwo,","))
        dataMapTwo <- data.frame( Value = character(),
        Name = character(),
        stringsAsFactors = F )

        for(x in 1:nrow(splitMapTwo)){
            xthSetTwo <- as.data.frame(strsplit(as.character(splitMapTwo[x,]),"="))
            dataMapTwo[x,] <- data.frame(Value = as.character(round(as.numeric(as.character(xthSetTwo[1,])),3) ), ## sig figs set to three to match what java is doing. NOTE: This is all for the sake of FlowSOM
            Name = as.character(xthSetTwo[2,]),
            stringsAsFactors = F)
        }

        newCategoricalTwo <- data.frame("xyz" = character(),
        stringsAsFactors = F)
        colnames(newCategoricalTwo) <- catParTwo

        for (i in 1:nrow(categoricalTwo)) {

            flag = 1
            for(j in 1:nrow(dataMapTwo)) {

                if(flag > 0){
                    placeholder = "UNCALLED_CAT"      ## This will allow us to check if the categorical was found
                    catSwapTwo = categoricalTwo[i,]         ## If that category isn't found, this will allow us to put something sensible there
                }

                if(categoricalTwo[i,] == dataMapTwo[j,1]){
                    placeholder = dataMapTwo[j,2]
                    flag = 0
                }

                if(flag > 0 && placeholder == "UNCALLED_CAT"){              ## Assuming that some round of the 'j' loop finds no categorical match, we will (for a moment) automatically call the category as unknown...
                    newCategoricalTwo[i,] = paste0(catParTwo,"_",catSwapTwo)
                } else if (flag == 0 && placeholder != "UNCALLED_CAT"){     ## ... However, assuming that we eventually find a matching categorical, we will name that cell's population correctly!
                    newCategoricalTwo[i,] <- placeholder
                }
            }
        }

        prePreData[catParTwo] <- newCategoricalTwo

    }else {

        catListTwo2 <- unique(categoricalTwo)

        dataMapTwo = data.frame(
        Value = as.numeric(),
        Name = as.character(),
        stringsAsFactors = F
        )

      anInt <- 1
      for(value in catListTwo2[[catParTwo]]){
          dataMapTwo[anInt,] <- data.frame(Value = as.numeric(value),
                                      Name = paste0(catParTwo,"-", as.character(value)),
                                      stringsAsFactors = F)
        anInt <- anInt + 1
      }

    }
}








  isCatParTwo <- T
  if (catPar == "") {
    writeLines("No categorical parameter's selected for annotation")
  } else {
    nonUnaryCat <- which(colnames(prePreData) %in% c(catPar))
    daCat <- subset(prePreData, select = nonUnaryCat)
    daCat <- data.frame(daCat[,1])
    prePreData <- subset(prePreData, select = -nonUnaryCat )
    colnames(daCat) <- gsub(" ", "", colnames(daCat))


    if (catParTwo == "") {
      writeLines("No categorical parameter's selected for annotation")
      isCatParTwo <- F
    } else {
      nonUnaryCatTwo <- which(colnames(prePreData) %in% c(catParTwo))
      daCatTwo <- subset(prePreData, select = nonUnaryCatTwo)
      daCatTwo <- data.frame(daCatTwo[,1])
      prePreData <- subset(prePreData, select = -nonUnaryCatTwo )
      colnames(daCatTwo) <- gsub(" ", "", colnames(daCatTwo))
    }

  }

  ## Calibrate Annotation Parameter
  #         minim <- min(daCat)
  #         maxim <- max(daCat)
  #         num.cluste <- nrow(unique(daCat))
  #         factor <- NULL
  #         if (min(daCat) == 0){
  #           factor <- (num.cluste-1)/maxim
  #         }else{
  #           factor <- num.cluste/maxim
  #         }
  #         calibratedCat <- round(daCat*factor)
  #         daCat <- calibratedCat

    featureAnnotationsFrame <- data.frame(colnames(prePreData))
    cellAnnotationsFrame <- data.frame(rownames(prePreData))

  preData <- t(as.matrix(prePreData))

  t.data <- t(preData)

  colnames(t.data) <- gsub(" ", "", colnames(t.data))

  colnames(t.data) <- paste(colnames(t.data),"  Gene Expression")

  data <- t(t.data)

  colnames(data) <- paste0("CellNamed_", colnames(data))

  sparse.data <- Matrix(data , sparse = T )
  ##cellIDs look good here...

  dir.create("sgdm")

  dear.dir <- paste0(getwd(),"/sgdm")

  writeMM(obj = sparse.data, file=paste0(getwd(),"/sgdm/matrix.mtx"))
  write.table(t(data[0,]), file=paste0(getwd(),"/sgdm/barcodes.tsv"), col.names=F, row.names=T, quote = F)
  write.table(data[,0], file=paste0(getwd(),"/sgdm/features.tsv"), col.names=F, row.names=T, quote = F)

  data.obj <- load10x(dear.dir,1)
  my.obj <- make.obj(data.obj)
#######################

  ## GET main.data BUILT ##
  #######################

    min.mito = 0
    max.mito = 1
    min.genes = 0
    max.genes = Inf
    min.umis = 0
    max.umis = Inf
    filter.by.cell.id = "character"
    keep.cell.id = "character"
    filter.by.gene = "character"
    filter.by.gene.exp.min = 1

    if ("iCellR" != class(my.obj)[1]) {
      stop("Your Object should be an object of class iCellR")
    }
    # take input from
    DATA <- my.obj@raw.data

    # filter by gene
    if (filter.by.gene[1] != "character"){
      genExist <- dim(subset(DATA, row.names(DATA) %in% filter.by.gene))[1]
      numberOfgenes <- length(filter.by.gene)
      if (genExist != numberOfgenes) {
        stop("your data lacks the gene/genes in filter.by.gene.")
      }
      gene.filt <- t(subset(DATA, row.names(DATA) %in% filter.by.gene) >= filter.by.gene.exp.min)
      # gene.filted <- as.data.frame(rowSums(gene.filt) > 0L)
      gene.filted <- as.data.frame(apply(gene.filt, 1, any))
      colnames(gene.filted) <- "gene.filter"
      gene.filt.cells <- row.names(subset(gene.filted, gene.filter == F))
      goneCells <- length(gene.filt.cells)
      allCells <- length(colnames(DATA))
    }

    GenesForFilter = paste(filter.by.gene, collapse=",")
    if (filter.by.gene[1] == "character") {
      GenesForFilter = "no cells were filtered based on expression."
    }


    ######## filter out ids
    if (filter.by.cell.id[1] != "character"){
      if (keep.cell.id[1] != "character"){
        stop("if filter.by.cell.id is set you can't set keep.cell.id")
      }
      z = filter.by.cell.id %in% colnames(DATA)
      Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
      Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
      if (length(filter.by.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
      }
      DATA <- DATA[ , -which(names(DATA) %in% filter.by.cell.id)]
      FiltCellIds <- paste(filter.by.cell.id , collapse=",")
      message(paste("The following cells were filtered out: ", FiltCellIds))
    }
    if (filter.by.cell.id[1] == "character") {
      FiltCellIds = "no cell ids were filtered."
    }


    ######## keep ids
    if (keep.cell.id[1] != "character"){
      if (filter.by.cell.id[1] != "character"){
        stop("if keep.cell.id is set you can't set filter.by.cell.id")
      }
      z = keep.cell.id %in% colnames(DATA)
      Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
      Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
      if (length(keep.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
      }
      DATA <- DATA[ , which(names(DATA) %in% keep.cell.id)]
      FiltCellIds <- paste(keep.cell.id, collapse=",")
      message(paste("The following cells are kept: ", FiltCellIds))
    }
    if (keep.cell.id[1] == "character") {
      FiltCellIds = "no cell ids to keep."
    }


    # make a filter parameters file
    FilterFile <- paste("Filters are:
- min mito",min.mito,"max mito",max.mito,"
- min # of genes",min.genes,"max # of genes",max.genes,"
- min UMIs",min.umis,"max UMIs",max.umis,"
- Cells were also filtered by the expression of the following genes (user input):
",GenesForFilter,"
- In addition the following cell ids were set to keep/filter (user input):
",FiltCellIds)
    # write filter parameters
    # write.table((FilterFile),file="filters_set.txt", row.names =F, quote = F, col.names = F)
    # message("filters_set.txt file has beed generated and includes the filters set for this experiment.")
    # return data
    row.names(DATA) <- gsub("-",".", row.names(DATA))
    attributes(my.obj)$main.data <- DATA
    attributes(my.obj)$my.filters <- FilterFile
#######################

 if(catPar != ""){
   renameMe <- paste0("CellNamed_", colnames(data))
   mainCat <- daCat
   colnames(mainCat) <- "clusters"
   mainCatNotZero <- subset(mainCat, clusters!=0)
   catList <- unique(mainCatNotZero)
         num.clust <- length(unique(mainCatNotZero[[1]]))
            Number_clusters <- num.clust
            names(Number_clusters) <- "Number_cluster"
            rownames(mainCat) <- colnames(data)
            clusters <- (mainCat[,1])
            res <- "All.index"
            partition <- as.list(mainCat[,1])
            names(partition) <- rownames(mainCat)
            resultats <- (as.list(Number_clusters))
            cluster.data <- list(All.index=res, Best.nc=resultats, Best.partition=partition)
            attributes(my.obj)$best.clust <- data.frame(clusters)
            attributes(my.obj)$cluster.data <- cluster.data
            my.obj@data.conditions <- "Data conditions: no conditions/single sample"
          clustInfo = as.data.frame(my.obj@cluster.data$Best.partition)
   }

   write.table(num.clust, paste0("Estimated_k_", runID), row.names = F, col.names = F)
   #######################


if(!isCatParTwo){
    num.clust.two = 0
}

   if(catParTwo != ""){
     # renameMe <- paste0("CellNamed_", colnames(data))
     mainCatTwo <- daCatTwo
     colnames(mainCatTwo) <- "condition"
     mainCatNotZeroTwo <- subset(mainCatTwo, clusters!=0)
     catListTwo <- unique(categoricalTwo)
     num.clust.two <- length(unique(mainCatNotZeroTwo[[1]]))
     Number_clusters_two <- num.clust.two
     names(Number_clusters_two) <- "Number_condition"
     rownames(mainCatTwo) <- colnames(data)
     condition <- (mainCatTwo[,1])
     resTwo <- "All.index"
     partitionTwo <- as.list(mainCatTwo[,1])
     names(partitionTwo) <- rownames(mainCatTwo)
     resultatsTwo <- (as.list(Number_clusters_two))
     cluster.data.two <- list(All.index=resTwo, Best.nc=resultatsTwo, Best.partition=partitionTwo)

     ## putting this into a slot of the data, but its sort of random... :/
     attributes(my.obj)$clust.cond.freq <- data.frame(condition)
     # attributes(my.obj)$best.clust <- data.frame(condition)

     attributes(my.obj)$metadata <- cluster.data.two
     # attributes(my.obj)$cluster.data <- cluster.data.two

     # my.obj@data.conditions <- cluster.data.two

     # clustInfo = as.data.frame(my.obj@cluster.data$Best.partition)
     clustInfoTwo = as.data.frame(my.obj@metadata$Best.partition )
   }

   write.table(num.clust.two, paste0("Estimated_conditionNumber_", runID), row.names = F, col.names = F)
   #######################




## DIFFERENTIAL EXPRESSION ##
#############################

DEG.list <- 0

UqValue <- fQVal
fcT <- fFC
fcTinv <- 1/fFC
fcTinv <- round( fcTinv, 4)

catList <- unlist(catList)

sortedDataMap <- dataMap[order(as.numeric(dataMap$Value)),]


if ( !isCatParTwo ){


    #### Create not gates for each cluster:
    if (num.clust <= 1){
        writeLines("No clusters selected, or clusters less than or equal to zero.")
    } else {
        # for(n in 1:num.clust){

        for(n in catList){

            nth <- which (catList == n)
            o <- n
            p <- c()
            # for (i in 1:num.clust){
            for (i in catList){
                if (i !=o){
                    p <- c(p,i)
                }
            }

            a <- table(my.obj@best.clust$clusters)
            a.num <- a[names(a)==o]


            if (a.num > 1){

                ##### Calculate DEG for one cluster:
                ## Typically this can be run in one line: (however something has gone wrong, so I'm running the algorithm myself)
                # diff.res <- run.diff.exp(my.obj, de.by = "clusters", cond.1 = c(n), cond.2 = p)

                ###########
                dat <- my.obj@main.data

                # 2 dimentions (is and isNot)
                DATA <- my.obj@best.clust

                ## set which clusters you want as condition 1 and 2
                de.by = "clusters"
                CondB = o
                CondA = p
                CondAnames = paste(CondA, collapse="_")
                CondBnames = paste(CondB, collapse="_")

                ## clusters
                # if (de.by == "clusters") {
                if (is.null(DATA$clusters)) {
                    stop("Clusters are not assigend yet, please run assign.clust fisrt.")
                } else {
                    Table=DATA
                    Cluster0 <- row.names(subset(Table, Table$clusters %in% CondA))
                    Cluster1 <- row.names(subset(Table, Table$clusters %in% CondB))
                }
                #  }

                ############## Filter
                datums <- data.frame(t(dat), row.names = NULL)
                cond1 <- data.frame(datums[Cluster0,], row.names = NULL) # , row.names = NULL)
                # meansCond1 <- sapply(cond1, mean, na.rm = T)

                cond2 <- data.frame(datums[Cluster1,]) #, rownames = "")

                ### merge both for pval length not matching error
                mrgd <- rbind(cond1,cond2)

                # mean
                meansCond1 <- sapply(cond1, mean, na.rm = T)
                meansCond2 <- sapply(cond2, mean, na.rm = T)
                baseMean <- sapply(mrgd, mean, na.rm = T)

                # FC
                initFC <- meansCond2/meansCond1
                FC <- (initFC)
                suppressWarnings( FC.log2 <- log2(FC) )

                # dims
                Cond1_Start <- 1
                Cond1_End <- dim(t(cond1))[2]
                Cond2_Start <- dim(t(cond1))[2] + 1
                Cond2_End <- dim(t(cond1))[2] + dim(t(cond2))[2]

                x = data.frame(mrgd[Cond1_Start:Cond1_End,], row.names = NULL)
                y = data.frame(mrgd[Cond2_Start:Cond2_End,], row.names = NULL)

                mrgdm <- data.frame(mrgd)
                mrgdm <- mrgd
                mrgdm <- t(mrgd)

                # indx <- !rowSums(new.CL[,1]==new.CL)==ncol(new.CL)
                indx <- !!rowSums(mrgd[,1]!=mrgd)
                new.mrgdm <- mrgd[indx,]

                Pval <- sapply(new.mrgdm, function(new.mrgdm) tryCatch ({
                    t.test( x = new.mrgdm[Cond1_Start:Cond1_End], y = new.mrgdm[Cond2_Start:Cond2_End] )$p.value
                }, error = function(e) {NA}))

                # padj
                FDR <- p.adjust(Pval)

                # combine
                Stats <- cbind(
                baseMean = baseMean,
                MeanCond1 = meansCond1,
                MeanCond2 = meansCond2,
                FC=FC,
                FC.log2=FC.log2,
                t.test=Pval,
                t.test.adj=FDR)

                # name column
                colnames(Stats) <- c("baseMean",CondAnames, CondBnames,"foldChange","log2FoldChange","pval","padj")

                ##### Write out DEG parameterSets as CSV files:
                DEG <- data.frame(featureAnnotationsFrame)

                FC <- data.frame(Stats[,"foldChange"])
                PV <- data.frame(Stats[,"padj"])

                DEG <- cbind(DEG, FC, PV)

                qValTH <- subset(DEG, (DEG$Stats....padj.. < UqValue))
                upFCnQVal <- subset(qValTH, qValTH$Stats....foldChange.. > fcT)
                downFCnQVal <- subset(qValTH, qValTH$Stats....foldChange.. < fcTinv)

                nam <- paste("Up_in", n, sep = "_")
                namAlt <- paste("Down_in", n, sep = "_")
                assign(nam, upFCnQVal)
                assign(namAlt, downFCnQVal)

                namGS <- paste0("/Up_in_", catParRaw, "_", nth, "_", runID, ".csv")
                fGSnam <- paste0(fjOuts,namGS)
                write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
                write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                write.table(paste0("$GeneSetName,",nam,"_RunID:_", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                write.table(paste0("$GeneSetDescription,Generated in the iCellR Pipeline. Fold-Change > ",fFC,"; q-Val (FDR) &lt; ", fQVal), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                write.table("[Genes],Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                write.table(upFCnQVal, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")

                namGSd <- paste0("/Down_in_", catParRaw, "_", nth, "_", runID, ".csv")
                fGSnamd <- paste0(fjOuts,namGSd)
                write.table("[GeneSet]", file = fGSnamd, append = F, quote = F, row.names = F, col.names = F )
                write.table("$SchemaVersion,1", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                write.table(paste0("$GeneSetName,", namAlt,"_RunID:_", runID), file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                write.table(paste0("$GeneSetDescription,Generated in the iCellR Pipeline. Fold-Change &lt; ", fcTinv, "; q-Val (FDR) &lt; ", fQVal), file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                write.table(" ", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                write.table("[Genes],Fold-Change,q-Value", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                write.table(downFCnQVal, file = fGSnamd, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")

            }else{
                writeLines(paste0("Cluster number ", n, " contains one or less events."))
            }
        }
    }

    #######################
    #######################

} else {

    ########################################
    ##### DEAL WITH SECOND CATEGORICAL #####
    ########################################

    sortedDataMapTwo <- dataMapTwo[order(as.numeric(dataMapTwo$Value)),]
    dat <- my.obj@main.data

    colnames(daCat) <- catPar
    tDat <- t(dat)
    newColNames <- gsub("(\\.\\.\\.Gene\\.Expression\\.)","",colnames(tDat))
    colnames(tDat)<- newColNames

    tester <- cbind(daCat,tDat, categoricalTwo)

    mth = 1
    for(clusterOfInterest in sortedDataMap[,2]){

        # writeLines(clusterOfInterest)

        # Create a new object (iCellR) containing only the cluster of interest cells
        if(hasPopMap){
            smallFrame <- tester[tester[[catPar]] == clusterOfInterest,]
        } else {
            thisValue <- sortedDataMap[sortedDataMap$Name == clusterOfInterest,]$Value
            smallFrame <- tester[tester[[catPar]] == thisValue, ]
        }


        #### Create not gates for each cluster:
        if (num.clust.two <= 1){
            writeLines("No clusters selected, or clusters less than or equal to zero.")
        } else {
            # for(n in 1:num.clust){


            catListTwo <- unlist(catListTwo)
            for(n in catListTwo){

                t <- ncol(smallFrame)

                nth <- which (catListTwo == n)
                o <- n
                p <- c()
                # for (i in 1:num.clust){
                for (i in catListTwo){
                    if (i !=o){
                        p <- c(p,i)
                    }
                }

                # a <- table(smallFrame[,t])
                # a.num <- a[names(a)==o]
                #
                #
                # if (a.num > 1){

                    ##### Calculate DEG for one cluster:
                    ## Typically this can be run in one line: (however something has gone wrong, so I'm running the algorithm myself)
                    # diff.res <- run.diff.exp(my.obj, de.by = "clusters", cond.1 = c(n), cond.2 = p)

                    ###########
                    tm <- t - 1
                    dat <- smallFrame[,2:tm]

                    # 2 dimentions (is and isNot)
                    DATA <- smallFrame[,t]

                    ## set which clusters you want as condition 1 and 2
                    de.by = "clusters"
                    CondB = o
                    CondA = p
                    CondAnames = paste(CondA, collapse="_")
                    CondBnames = paste(CondB, collapse="_")

                    ## clusters
                    # if (de.by == "clusters") {
                    if (is.null(DATA)) {
                        stop("Clusters are not assigend yet, please run assign.clust fisrt.")
                    } else {
                        # Table=DATA

                        suppressWarnings( Cluster0 <- smallFrame[smallFrame[[catParTwoRaw]]==CondA,] )
                        # Cluster0 <- smallFrame[smallFrame[[catParTwoRaw]]==CondAnames,] #subset(Table==1 )
                        if(nrow(Cluster0) == 0){
                          Cluster0 <- smallFrame[smallFrame[[catParTwo]]==CondAnames,]
                        }

                        Cluster1 <- smallFrame[smallFrame[[catParTwoRaw]]==CondBnames,]
                        if(nrow(Cluster1) == 0){
                          Cluster1 <- smallFrame[smallFrame[[catParTwo]]==CondBnames,]
                        }

                        cond1 <- Cluster0[,2:tm]
                        cond2 <- Cluster1[,2:tm]

                    }

                    ############## Filter
                    # datums <- data.frame(t(dat), row.names = NULL)
                    # cond1 <- data.frame(datums[Cluster0,], row.names = NULL) # , row.names = NULL)
                    # cond1 <- data.frame(datums[Cluster0,], row.names = NULL) # , row.names = NULL)
                    # meansCond1 <- sapply(cond1, mean, na.rm = T)

                    # cond2 <- data.frame(datums[Cluster1,]) #, rownames = "")

                    ### merge both for pval length not matching error
                    mrgd <- rbind(cond1,cond2)

                    # mean
                    meansCond1 <- sapply(cond1, mean, na.rm = T)
                    meansCond2 <- sapply(cond2, mean, na.rm = T)
                    baseMean <- sapply(mrgd, mean, na.rm = T)

                    # FC
                    initFC <- meansCond2/meansCond1
                    FC <- (initFC)
                    suppressWarnings( FC.log2 <- log2(FC) )

                    # dims
                    Cond1_Start <- 1
                    Cond1_End <- dim(t(cond1))[2]
                    Cond2_Start <- dim(t(cond1))[2] + 1
                    Cond2_End <- dim(t(cond1))[2] + dim(t(cond2))[2]

                    x = data.frame(mrgd[Cond1_Start:Cond1_End,], row.names = NULL)
                    y = data.frame(mrgd[Cond2_Start:Cond2_End,], row.names = NULL)

                    mrgdm <- data.frame(mrgd)
                    mrgdm <- mrgd
                    mrgdm <- t(mrgd)

                    # indx <- !rowSums(new.CL[,1]==new.CL)==ncol(new.CL)
                    indx <- !!rowSums(mrgd[,1]!=mrgd)
                    new.mrgdm <- mrgd[indx,]

                    Pval <- sapply(new.mrgdm, function(new.mrgdm) tryCatch ({
                        t.test( x = new.mrgdm[Cond1_Start:Cond1_End], y = new.mrgdm[Cond2_Start:Cond2_End] )$p.value
                    }, error = function(e) {NA}))

                    # padj
                    FDR <- p.adjust(Pval)

                    # combine
                    Stats <- cbind(
                    baseMean = baseMean,
                    MeanCond1 = meansCond1,
                    MeanCond2 = meansCond2,
                    FC=FC,
                    FC.log2=FC.log2,
                    t.test=Pval,
                    t.test.adj=FDR)

                    # name column
                    colnames(Stats) <- c("baseMean",CondAnames, CondBnames,"foldChange","log2FoldChange","pval","padj")

                    ##### Write out DEG parameterSets as CSV files:
                    DEG <- data.frame(featureAnnotationsFrame)

                    FC <- data.frame(Stats[,"foldChange"])
                    PV <- data.frame(Stats[,"padj"])

                    DEG <- cbind(DEG, FC, PV)

                    qValTH <- subset(DEG, (DEG$Stats....padj.. < UqValue))
                    upFCnQVal <- subset(qValTH, qValTH$Stats....foldChange.. > fcT)
                    downFCnQVal <- subset(qValTH, qValTH$Stats....foldChange.. < fcTinv)

                    nam <- paste("Up_in", dataMapTwo[n,2],clusterOfInterest, sep = "_")       # The dataMapTwo here is coming up wrong!
                    namAlt <- paste("Down_in", dataMapTwo[n,2],clusterOfInterest, sep = "_")  # The dataMapTwo here is coming up with the wrong category when secondary cat parameter is ambiguous
                    assign(nam, upFCnQVal)
                    assign(namAlt, downFCnQVal)

                    namGS <- paste0("/Up_in_",catParTwoRaw,"_",nth, "_", catParRaw, "_",mth,"_", runID, ".csv")
                    fGSnam <- paste0(fjOuts,namGS)
                    write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
                    write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                    write.table(paste0("$GeneSetName,",nam,"_RunID:_", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                    write.table(paste0("$GeneSetDescription,Generated in the iCellR Pipeline. Fold-Change > ",fFC,"; q-Val (FDR) &lt; ", fQVal), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                    write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                    write.table("[Genes],Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
                    write.table(upFCnQVal, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")

                    namGSd <- paste0("/Down_in_", catParTwoRaw, "_", nth, "_",catParRaw,"_", mth,"_", runID, ".csv")
                    fGSnamd <- paste0(fjOuts,namGSd)
                    write.table("[GeneSet]", file = fGSnamd, append = F, quote = F, row.names = F, col.names = F )
                    write.table("$SchemaVersion,1", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                    write.table(paste0("$GeneSetName,", namAlt,"_RunID:_", runID), file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                    write.table(paste0("$GeneSetDescription,Generated in the iCellR Pipeline. Fold-Change &lt; ", fcTinv, "; q-Val (FDR) &lt; ", fQVal), file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                    write.table(" ", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                    write.table("[Genes],Fold-Change,q-Value", file = fGSnamd, append = T, quote = F, row.names = F, col.names = F )
                    write.table(downFCnQVal, file = fGSnamd, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")

                # }else{
                #     writeLines(paste0("Cluster ", n, " contains one or less events."))
                # }
            }
        }

        mth = mth + 1

    }

}
#################
### SAVE GAME ###
save(my.obj, file = "my.obj.Robj")
#################

write.table(data.frame(1,2,3,5,8), file="fin.tsv", sep="  ", col.names=F, row.names = F)