
## DESCRIPTION ##
#######################################################################
# Copyright (c) 2019 Ian Taylor, Becton Dickinson and Company
#
# Script based on package developed by rezakj, iCellR: https://github.com/rezakj/iCellR
# Plugin for SeqGeq and FlowJo by: Ian Taylor, Josh Luthy, Dylan Hinson, Shahid Siddiq,
# Matt Swindle, and Richard Halpert of BD Life Sciences, Ashland, OR
#
# License
# The software is distributed under the terms of the
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
#
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability, maintenance and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this
# software.
#######################################################################

## INSTALL REQUIRED PACKAGES ##
#######################
# Install from CRAN

# These are the dependancies to use!!:
# install.packages("iCellR")
# (Required for CellCycle!:)
# install.packages("phateR")
# install.packages("Matrix")

## TRY TO AUTO INSTALL ##

# Install function for packages
packages<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="http://cran.r-project.org")
    require(x,character.only=TRUE)
  }
}

# the Real Deal (aka required packages) #
packages(backports)
packages(iCellR)
packages(Matrix)
packages(igraph)
#######################

## LOAD INFO FROM FJ/SG ##
#######################

EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT

runKNETL <- FJ_KNETL
# doClust <- F
doClust <- FJ_ICLUST
doPCA <- FJ_PCA
doTSNE <- FJ_TSNE

parNames <- c(FJ_PARAMS_LIST)
zoomZoom <- FJ_ZOOM
fjOuts <- "FJ_OUTS"
fjDataFilePath <- "FJ_DATA_FILE_PATH"
fjCsvOutputFile <- "FJ_CSV_OUPUT_FILE"
runID <- "FJ_RUNID"
fjDerParName <- "FJ_PAR_DER_PAR_NAME"
# fjK <- FJ_PAR_K_PAR
isSeqGeq <- FJ_ISSEQGEQ
# fQVal <- FJ_fQVal
# fFC <- FJ_FC
catPar <- ""
catPar <- gsub("(\\[)","",gsub("(\\])","",catPar))

clustMtx <- ""
# "ward.D", "ward.D2", "single", "complete", "average", "mcquitty", "median", "centroid", "kmeans", "NONE"

clustDist <- ""
# "euclidean", "maximum", "manhattan", "canberra", "binary", "minkowski" or "NULL"

clustIdx <- ""
# "kl", "ch", "hartigan", "ccc", "scott", "marriot", "trcovw", "tracew", "friedman", "rubin",
# "cindex", "db", "silhouette", "duda", "pseudot2", "beale", "ratkowsky", "ball", "ptbiserial",
# "gap", "frey", "mcclain", "gamma", "gplus", "tau", "dunn", "hubert", "sdindex", "dindex", "sdbw"


if (clustMtx == "NONE") {
  colorBy = "monochrome"
}else{
  colorBy = "clusters"
}

## set your working directory
setwd(fjOuts)
#######################

## FUNCTIONS ##
#######################
## For Error Messaging
stopAndReturnError <- function(message, ...) {
    cat(message, file=paste0(fjCsvOutputFile, "_err.txt"))
    stop(message, ...)
}



findMarkersNew <- function (x = NULL,
data.type = "main",
fold.change = 2,
padjval = 0.1,
Inf.FCs = FALSE,
uniq = FALSE,
positive = TRUE) {
    if ("iCellR" != class(x)[1]) {
        stop("x should be an object of class iCellR")
    }

    x = my.obj
    data.type = "main"
    fold.change = 2
    padjval = 0.1
    Inf.FCs = FALSE
    uniq = FALSE
    positive = TRUE

    ###########
    ## get main data
    if (data.type == "main") {
        dat <- x@main.data
    }
    if (data.type == "imputed") {
        dat <- x@imputed.data
    }
    # get cluster data
    # get avrages
    x <- clust.avg.exp(x, data.type = "imputed")
    DATA <- x@best.clust
    ############## set wich clusters you want as condition 1 and 2
    MyClusts <- as.numeric(unique(DATA$clusters))
    MyClusts <- sort(MyClusts)
    ############################### loop start
    for (i in MyClusts) {
        message(paste(" Finding markers for cluster:",i,"..."))
        Noi <- MyClusts[-which((MyClusts) %in% i)]
        Table=DATA
        Cluster0 <- row.names(subset(Table, Table$clusters %in% i))
        Cluster1 <- row.names(subset(Table, Table$clusters %in% Noi))
        ############## Filter
        #    cond1 <- dat[,Cluster0]
        #    cond2 <- dat[,Cluster1]
        cond1 <- dat[ , which(names(dat) %in% Cluster0)]
        cond2 <- dat[ , which(names(dat) %in% Cluster1)]
        ### merge both for pval length not matching error
        mrgd <- cbind(cond1,cond2)
        #    mrgd <- merge(cond1, cond2, by="row.names")
        #    row.names(mrgd) <- mrgd$Row.names
        #    mrgd <- mrgd[,-1]
        mrgd <- data.matrix(mrgd)
        # mean
        meansCond1 <- apply(cond1, 1, function(cond1) {mean(cond1)})
        meansCond2 <- apply(cond2, 1, function(cond2) {mean(cond2)})
        baseMean <- apply(mrgd, 1, function(mrgd) {mean(mrgd)})
        baseSD <- apply(mrgd, 1, function(mrgd) {sd(mrgd)})
        # FC
        FC <- meansCond1/meansCond2
        FC.log2 <- log2(FC)
        # dims
        Cond1_Start <- 1
        Cond1_End <- dim(cond1)[2]
        Cond2_Start <- dim(cond1)[2] + 1
        Cond2_End <- dim(cond1)[2] + dim(cond2)[2]
        # filter
        if (positive == TRUE) {
            FiltData <- subset(FC.log2, FC.log2 > log2(fold.change))
        } else {
            FiltData <- subset(FC.log2, FC.log2 < -log2(fold.change) | FC.log2 > log2(fold.change))
        }
        mrgd <- subset(mrgd, row.names(mrgd) %in% row.names(as.data.frame(FiltData)))
        # pval
        Pval <- apply(mrgd, 1, function(mrgd) {
            t.test(x = mrgd[Cond1_Start:Cond1_End], y = mrgd[Cond2_Start:Cond2_End])$p.value
        })
        # padj
        FDR <- p.adjust(Pval)
        # combine
        Stats <- cbind(
        baseMean = baseMean,
        baseSD = baseSD,
        MeanCond1 = meansCond1,
        MeanCond2 = meansCond2,
        FC=FC,
        FC.log2=FC.log2)
        colnames(Stats) <- c("baseMean","baseSD","AvExpInCluster", "AvExpInOtherClusters","foldChange","log2FoldChange")
        # cbind pvals
        clusters <- rep(i,length(FDR))
        # make cluster names
        Stats1 <- cbind(
        pval = Pval,
        padj = FDR,
        clusters = clusters)
        # filter
        Stats <- as.data.frame(Stats)
        Stats1 <- as.data.frame(Stats1)
        Stats1 <- subset(Stats1, Stats1$padj < padjval)
        # merge both pvals and stats
        mrgdall <- merge(Stats, Stats1, by="row.names")
        row.names(mrgdall) <- mrgdall$Row.names
        mrgdall <- mrgdall[,-1]
        ############################# get avrage data
        #    if (add.avg == TRUE) {
        AvData <- x@clust.avg
        row.names(AvData) <- AvData$gene
        mrgdall <- merge(mrgdall, AvData, by="row.names")
        row.names(mrgdall) <- mrgdall$Row.names
        mrgdall <- mrgdall[,-1]
        #    }
        # make it an object
        DatNmaes=paste("DATAcluster",i,sep="_")
        eval(call("<-", as.name(DatNmaes), mrgdall))
    }
    ############################### loop end
    # merge all
    filenames <- ls(pattern="DATAcluster_")
    datalist <- mget(filenames)
    df <- do.call("rbind", datalist)
    row.names(df) <- make.names(df$gene, unique=TRUE)
    df <- subset(df, gene != "NA")
    ####
    if (uniq == TRUE) {
        data <- (as.data.frame(table(df$gene)))
        datanew <- (data[order(data$Freq, decreasing = TRUE),])
        datanew <- subset(datanew, datanew$Freq == 1)
        datanew <- as.character(datanew$Var1)
        myDATA = df
        myDATA <- subset(myDATA, myDATA$gene %in% datanew)
        df <- myDATA
    }
    if (Inf.FCs == FALSE) {
        df <- subset(df, log2FoldChange != Inf)
        df <- subset(df, log2FoldChange != Inf & log2FoldChange != -Inf)
    }
    df <- df[order(df$log2FoldChange,decreasing = TRUE),]
    df <- df[order(df$clusters,decreasing = FALSE),]
    message("All done!")
    return(df)
}



pseudotime.knetl <- function (x = NULL,
dist.method = "euclidean",
k = 5,
abstract = TRUE,
data.type = "pca",
dims = 1:20,
conds.to.plot = NULL,
my.layout = "layout_with_fr",
node.size = 10,
cluster.membership = FALSE,
interactive = TRUE,
node.colors = NULL,
edge.color = "gray",
out.name = "Pseudotime.Abstract.KNetL",
my.seed = 1
) {
    #
    start_time1 <- Sys.time()
    #  cluster
    if(data.type == "pca") {
        DATA = (t(x@pca.data))[dims, ]
        message(paste("Getting PCA data"))
    }
    if(data.type == "umap") {
        DATA <- t(x@umap.data[1:2])
        message(paste("Getting UMAP data"))
    }
    if(data.type == "tsne") {
        DATA <- t(x@tsne.data)
        message(paste("Getting tSNE data"))
    }
    #####
    MYtitle = "Pseudotime Abstract KNetL (PAK)"
    MyClusters <- x@best.clust
    if (!is.null(conds.to.plot)) {
        Conditions <- data.frame(do.call('rbind', strsplit(as.character(colnames(DATA)),'_',fixed=TRUE)))[1]
        Conditions <- as.character(as.matrix(Conditions))
        Conditions <- as.data.frame(cbind(colnames(DATA),Conditions))
        colnames(Conditions) <- c("row","conds")
        Conditions <- subset(Conditions, Conditions$conds %in% conds.to.plot)
        Conditions <- as.character(Conditions$row)
        DATA <- as.data.frame(t(DATA))
        DATA <- subset(DATA, rownames(DATA) %in% Conditions)
        DATA <- t(DATA)
        MYtitle = paste(MYtitle," (",conds.to.plot,")", sep="")
        MyClusters <- subset(MyClusters, rownames(MyClusters) %in% Conditions)
    }
    message(paste("   Calculating", dist.method,"distance ..."))
    My.distances = as.matrix(dist(t(DATA),dist.method))
    #####
    #####
    ncells=dim(DATA)[2]
    cell.num = k
    #####  ### time
    # pb <- progress_bar$new(total = ncells,
    #                        format = "[:bar] :current/:total (:percent) :elapsedfull eta: :eta",
    #                        clear = FALSE, width= 60)
    message(paste("    Finding",cell.num, "neighboring cells per cell ..."))
    KNN1 = lapply(1:ncells, function(findKNN){
        # pb$tick()
        #  order(My.distances[,findKNN])[2:cell.num]})
        #  MyOrd <- GETord(My.distances[,findKNN])[2:cell.num]
        MyOrd <- order(My.distances[,findKNN])[2:cell.num]
        MyOrd.IDs <- row.names(My.distances)[MyOrd]
        MyOrd.IDs.clust <- as.numeric(as.matrix(subset(MyClusters, row.names(MyClusters) %in% MyOrd.IDs)))
        ####
        MyDist <- My.distances[MyOrd]
        ####
        MyRoot <- rep(findKNN,cell.num-1)
        MYRoot.ID <- colnames(My.distances)[findKNN]
        MYRoot.IDs <- rep(MYRoot.ID,length(MyRoot))
        MYRoot.clust <- as.numeric(as.matrix(subset(MyClusters, row.names(MyClusters) %in% MYRoot.ID)))
        MYRoot.clusts <- rep(MYRoot.clust,length(MyRoot))
        ####
        data <- cbind(MyRoot,MyOrd,MyDist,MYRoot.IDs,MyOrd.IDs,MYRoot.clusts,MyOrd.IDs.clust)
        colnames(data)<- c("from","to","weight","from.id","to.id","from.clust","to.clust")
        data <- as.data.frame(data)
    })
    ########
    if (abstract == TRUE) {
        message("    Generating abstract graph ...")
        data <- do.call("rbind", KNN1)
        df <- as.data.frame(table(data[6:7]))
        colnames(df) <- c("from","to","weight")
        #    df <- subset(df, df$weight != 0)
        df <- subset(df, df$weight > 5)
        gg <- graph.data.frame(df, directed=FALSE)
        message("   Simplifying graph ...")
        gg <- simplify(gg)
    }
    #######
    if (abstract == FALSE) {
        message("    Generating graph ...")
        data <- do.call("rbind", KNN1)
        df <- as.data.frame(data[1:3])
        colnames(df) <- c("from","to","weight")
        gg <- graph.data.frame(df, directed=FALSE)
    }
    ### make colors
    ####################
    ######
    if (is.null(node.colors)) {
        gg_color_hue <- function(n) {
            hues = seq(15, 375, length = n + 1)
            hcl(h = hues, l = 65, c = 100)[1:n]
        }
        n = max(MyClusters$clusters)# +1
        colbar = gg_color_hue(n)
    } else {
        colbar = node.colors
    }
    if (abstract == FALSE) {
        colbar <- colbar[MyClusters$clusters]
    }
    ###########
    #    V(gg)$color <- colbar
    # barplot(c(1:11), col=colbar)
    ####
    MYedgeWith <- log2(as.numeric(E(gg)$weight) + 1)

    ## https://briatte.github.io/ggnet/
    message("Drawing layout and generating plot ...")
    ###############

    rm(.Random.seed, envir=.GlobalEnv)
    set.seed(my.seed)
    if (cluster.membership == TRUE) {
        ceb <- cluster_fast_greedy(as.undirected(gg))
        return(
        plot(ceb,gg, layout=get(my.layout),
        vertex.size=node.size,
        vertex.color=colbar,
        edge.width=MYedgeWith,
        main = MYtitle,
        edge.color=edge.color))
    }
    #######
    if (cluster.membership == FALSE) {
        if (abstract == FALSE) {
            return(
            plot(gg, layout=layout_nicely,
            vertex.size=(node.size/2),
            vertex.label=NA,
            margin=0,
            vertex.color=colbar))
        }
        return(
        plot(gg, layout=get(my.layout),
        vertex.size=node.size,
        vertex.color=colbar,
        edge.width=MYedgeWith,
        main = MYtitle,
        edge.color=edge.color))
    }
}





## LOAD DATA ##
#######################
library(Matrix)
library(iCellR)

## TODO UNCOMMENT THIS OUT
sample <- fjDataFilePath

## FOR TESTING ONLY ##
# setwd("/Users/10201397/Desktop/TODO/icellr_testing")
# sample <- "/Users/10201397/Desktop/TODO/icellr_testing/exportModTest.csv"

if (isSeqGeq){
    prePreData <- read.table(file= sample, sep = ",", row.names = 1, header = TRUE, check.names = FALSE)
} else {
    data <- read.csv(fjDataFilePath, check.names=FALSE)
    cellIdsColumn <- rep(NA, nrow(data))
    cellIds <- data.frame(CellIdDP = cellIdsColumn)
    names(cellIds) <- CellIdDP
    if (CellIdDP %in% names(data)) {
        cellIds <- data[, CellIdDP]
        data <- data[, -which(names(data) %in% c(CellIdDP))]
    }
    EventNumberColumn <- data$EventNumberDP
    eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
    # EventNumberColumn <- rep(NA, nrow(data))
    rownames(data) <- data$EventNumberDP
    prePreData <- data
    colnames(prePreData) <- c(parNames, EventNumberDP)
}

if (catPar == "") {
    writeLines("No categorical parameter's selected for annotation")
} else {
    if (isSeqGeq) {
        nonUnaryCat <- which(colnames(prePreData) %in% c(catPar))
        daCat <- subset(prePreData, select = nonUnaryCat)
        daCat <- data.frame(daCat[,1])
        prePreData <- subset(prePreData, select = -nonUnaryCat )
        colnames(daCat) <- gsub(" ", "", colnames(daCat))
    } else {
        nonUnaryCat <- which(colnames(prePreData) %in% c(catPar))
        daCat <- subset(prePreData, select = nonUnaryCat)
        daCat <- data.frame(daCat[,1])
        prePreData <- subset(prePreData, select = -nonUnaryCat )
        colnames(daCat) <- gsub(" ", "", colnames(daCat))
    }

}

featureAnnotationsFrame <- data.frame(colnames(prePreData))
cellAnnotationsFrame <- data.frame(rownames(prePreData))

preData <- t(as.matrix(prePreData))

t.data <- t(preData)
# t.data <- t(prePreData)

if (!isSeqGeq) {
    colnames(t.data) <- gsub(".*? :: (.+)", "\\1",colnames(t.data))
}

colnames(t.data) <- gsub(" ", "", colnames(t.data))

colnames(t.data) <- paste(colnames(t.data),"  Gene Expression")

data <- t(t.data)

if (catPar == ""){
    colnames(data) <- paste0("CellNamed_", colnames(data))
}else{


    ## Calibrate Annotation Parameter
    minim <- min(daCat)
    maxim <- max(daCat)
    num.cluste <- nrow(unique(daCat))
    factor <- NULL
    if (min(daCat) == 0){
        factor <- (num.cluste-1)/maxim
    }else{
        factor <- num.cluste/maxim
    }
    calibratedCat <- round(daCat*factor)
    daCat <- calibratedCat

    ## Add Annotations to CellIDs
    cellAnnoMod <- cellAnnotationsFrame
    concatMe <- cbind(daCat, cellAnnoMod)
    renameMe <- paste0(catPar, "=", concatMe[,1], "_CellNamed.", concatMe[,2])
    colnames(data) <- renameMe
}

sparse.data <- Matrix(data , sparse = T )
##cellIDs look good here...

dir.create("sgdm")

dear.dir <- paste0(getwd(),"/sgdm")

writeMM(obj = sparse.data, file=paste0(getwd(),"/sgdm/matrix.mtx"))
write.table(t(data[0,]), file=paste0(getwd(),"/sgdm/barcodes.tsv"), col.names=F, row.names=T, quote = F)
write.table(data[,0], file=paste0(getwd(),"/sgdm/features.tsv"), col.names=F, row.names=T, quote = F)

data.obj <- load10x(dear.dir,1)
my.obj <- make.obj(data.obj)
#######################

## GET main.data BUILT ##
#######################

min.mito = 0
max.mito = 1
min.genes = 0
max.genes = Inf
min.umis = 0
max.umis = Inf
filter.by.cell.id = "character"
keep.cell.id = "character"
filter.by.gene = "character"
filter.by.gene.exp.min = 1

if ("iCellR" != class(my.obj)[1]) {
    stop("Your Object should be an object of class iCellR")
}
# take input from
DATA <- my.obj@raw.data

# filter by gene
if (filter.by.gene[1] != "character"){
    genExist <- dim(subset(DATA, row.names(DATA) %in% filter.by.gene))[1]
    numberOfgenes <- length(filter.by.gene)
    if (genExist != numberOfgenes) {
        stop("your data lacks the gene/genes in filter.by.gene.")
    }
    gene.filt <- t(subset(DATA, row.names(DATA) %in% filter.by.gene) >= filter.by.gene.exp.min)
    # gene.filted <- as.data.frame(rowSums(gene.filt) > 0L)
    gene.filted <- as.data.frame(apply(gene.filt, 1, any))
    colnames(gene.filted) <- "gene.filter"
    gene.filt.cells <- row.names(subset(gene.filted, gene.filter == F))
    goneCells <- length(gene.filt.cells)
    allCells <- length(colnames(DATA))
}
#
GenesForFilter = paste(filter.by.gene, collapse=",")
if (filter.by.gene[1] == "character") {
    GenesForFilter = "no cells were filtered based on expression."
}

######## filter out ids
if (filter.by.cell.id[1] != "character"){
    if (keep.cell.id[1] != "character"){
        stop("if filter.by.cell.id is set you can't set keep.cell.id")
    }
    z = filter.by.cell.id %in% colnames(DATA)
    Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
    Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
    if (length(filter.by.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
    }
    DATA <- DATA[ , -which(names(DATA) %in% filter.by.cell.id)]
    FiltCellIds <- paste(filter.by.cell.id , collapse=",")
    message(paste("The following cells were filtered out: ", FiltCellIds))
}
if (filter.by.cell.id[1] == "character") {
    FiltCellIds = "no cell ids were filtered."
}
######## keep ids
if (keep.cell.id[1] != "character"){
    if (filter.by.cell.id[1] != "character"){
        stop("if keep.cell.id is set you can't set filter.by.cell.id")
    }
    z = keep.cell.id %in% colnames(DATA)
    Missing.cell.ids <- filter.by.cell.id[-which(z, arr.ind = T, useNames = T)]
    Missing.cell.ids <- paste(Missing.cell.ids, collapse=",")
    if (length(keep.cell.id) != length(z[z==TRUE])) {
        stop(paste("Your dataset lacks the following ids: ",Missing.cell.ids))
    }
    DATA <- DATA[ , which(names(DATA) %in% keep.cell.id)]
    FiltCellIds <- paste(keep.cell.id, collapse=",")
    message(paste("The following cells are kept: ", FiltCellIds))
}
if (keep.cell.id[1] == "character") {
    FiltCellIds = "no cell ids to keep."
}

# make a filter parameters file
FilterFile <- paste("Filters are:
- min mito",min.mito,"max mito",max.mito,"
- min # of genes",min.genes,"max # of genes",max.genes,"
- min UMIs",min.umis,"max UMIs",max.umis,"
- Cells were also filtered by the expression of the following genes (user input):
",GenesForFilter,"
- In addition the following cell ids were set to keep/filter (user input):
",FiltCellIds)

# return data
row.names(DATA) <- gsub("-",".", row.names(DATA))
attributes(my.obj)$main.data <- DATA
attributes(my.obj)$my.filters <- FilterFile
#######################


## PCA ##
#######################
if(doPCA){
    data.type = "pca"
    method = "base.mean.rank"
    top.rank = 500
    plus.log.value = 0.1
    batch.norm = FALSE
    gene.list = "character"

    raw.data.order <- my.obj@raw.data[ order(rowMeans(my.obj@raw.data), decreasing = TRUE), ]

    if (isSeqGeq){
        topGenes <- (raw.data.order)
        TopNormLogScale <- log(topGenes + plus.log.value)
    } else {
        topGenes <- (raw.data.order)
        sup <- min(topGenes)
        dawg <- topGenes + abs(sup)
        TopNormLogScale <- log(dawg + plus.log.value)
    }

    counts.pca <- prcomp(TopNormLogScale, center = FALSE, scale. = FALSE)
    attributes(my.obj@raw.data)$pca.info <- counts.pca
    # DATA
    dataPCA = data.frame(counts.pca$rotation) # [1:max.dim]
    attributes(my.obj@main.data)$pca.data <- dataPCA
    # optimal
    pcaDATA <- counts.pca$sdev
    OPTpcs <- mean(pcaDATA)*2
    OPTpcs <- (pcaDATA > OPTpcs)
    OPTpcs <- length(OPTpcs[OPTpcs==TRUE]) + 1
    attributes(my.obj@main.data)$opt.pcs <- OPTpcs
    attributes(my.obj)$pca.data <- dataPCA
    row.names(dataPCA) <- gsub("-",".", row.names(dataPCA))
    attributes(my.obj)$pca.data <- t(dataPCA)

    if (length(dataPCA) < 10) {
        dimens <- length(dataPCA)
    } else {
        dimens <- 10
    }
} else {
    ## What are the dimensions?:
    dimens <- nrow(my.obj@main.data)
    aThing <- my.obj@main.data

    my.obj@pca.data <- as.data.frame(aThing)
    # my.obj@pca.data <- as.data.frame(t(aThing))
    my.obj@pca.data$EventNumberDP...Gene.Expression. <- NULL
}

if(runKNETL){

    nDim <- ncol(my.obj@pca.data)

    x = my.obj
    dist.method = "euclidean"
    k = zoomZoom
    data.type = "pca"
    dims = 1:nDim
    joint = FALSE
    col.by = "clusters"
    my.seed = 1
    layout.2d = "layout_nicely"
    layout.3d = "layout_with_fr"
    add.3d = FALSE
    dim.redux = "umap"
    do.redux = TRUE
    run.iclust = doClust
    return.graph = FALSE

    start_time1 <- Sys.time()
    #  cluster
    if(data.type == "pca") {
        DATA = x@pca.data
        message(paste("Getting PCA data"))
    }

    #####
    message(paste("   Calculating", dist.method,"distance ..."))
    My.distances = as.matrix(dist(t(DATA),dist.method))
    #####
    #####
    ncells=dim(DATA)[2]
    cell.num = k
    #####

    if (joint == FALSE) {
        message(paste("    Finding",cell.num, "neighboring cells per cell ..."))
        KNN1 = lapply(1:ncells, function(findKNN){

            MyOrd <- order(My.distances[,findKNN])[2:cell.num]
            MyDist <- My.distances[MyOrd]
            MyRoot <- rep(findKNN,cell.num-1)
            data <- cbind(MyRoot,MyOrd,MyDist)
            colnames(data)<- c("from","to","weight")
            data <- as.data.frame(data)
        })
    }
    ###################
    if (joint == TRUE) {
        message(paste("    Finding",cell.num, "jointly neighboring cells per cell ..."))
        MyDF <- as.data.frame(row.names(My.distances))
        colnames(MyDF) <- "IDs"
        KNN1 = lapply(1:ncells, function(findKNN){
            ###### time
            pb$tick()
            #### end time
            ######### loop
            for(i in conditions){
                ha <- paste("^",i,"_",sep="")
                #      CellOrd <- colnames(My.distances)[(GETord(My.distances[,findKNN]))]
                CellOrd <- colnames(My.distances)[(order(My.distances[,findKNN]))]
                CellsId <- subset(CellOrd, grepl(ha, CellOrd))[1:cell.num]
                MyOrd <- as.numeric(rownames(subset(MyDF,MyDF$IDs %in% CellsId)))
                MyDist <- My.distances[MyOrd]
                MyRoot <- rep(findKNN,cell.num)
                data <- cbind(MyRoot,MyOrd,MyDist)
                colnames(data)<- c("from","to","weight")
                MYData <- as.data.frame(data)
                #          CellsId <- grep(ha,CellOrd,value=T, invert=F)[1:myNN]
                NameCol=paste("MySet",i,sep="_")
                eval(call("<-", as.name(NameCol), MYData))
            }
            filenames <- ls(pattern="MySet_")
            TheData <- do.call('rbind', mget(filenames))
            rownames(TheData) <- NULL
            TheData
        })
    }
    ##############
    data <- do.call("rbind", KNN1)
    #####
    data <- do.call("rbind", KNN1)
    message(paste("Generating graph from root to neighboring cells ..."))
    g <- igraph::graph.data.frame(data, directed=FALSE)

    #########
    if (return.graph == FALSE){
        message("Generating 2D Layouts ...")
        set.seed(my.seed)

        data2 <- get(layout.2d)(g) ## Not something we find in this... wtf?

        colnames(data2) <- c("V1","V2")

        data2 <- as.data.frame(scale(data2))
        if (add.3d == TRUE) {
            message("Generating 3D Layouts ...")
            data3 <- get(layout.3d)(g, dim =3)
            colnames(data3) <- c("V1","V2","V3")
            data3 <- as.data.frame(scale(data3))
            row.names(data3) <- colnames(DATA)
        }
        ########
        y <- x
        y@pca.data <- data2
        #######
        if (do.redux == TRUE) {
            if (dim.redux == "umap") {
                knn <- 15
                if(nrow(y@pca.data) < knn ){
                    knn <- nrow(y@pca.data)
                }
                message("Running Dimensionality Reduction (UMAP) ...")
                y <- run.umap(y, dims = 1:2, n_neighbors = knn)
                data2 <- y@umap.data
            }
        }
        ##################################
        #######
        end_time1 <- Sys.time()
        Time = difftime(end_time1,start_time1,units = "mins")
        Time = round(as.numeric(Time),digits = 2)
        message(paste("Total time to complete KNETL",Time,"mins"))
        attributes(x)$knetl.data <- data2
        if (add.3d == TRUE) {
            attributes(x)$knetl.data.3d <- data3
        }

        if (run.iclust == TRUE) {
            message("Running clustering ...")
            # y <- iclust(y, k = k, dims = 1:2)

            dist.method = "euclidean"
            k = k
            data.type = "pca"
            dims = 1:2
            return.graph = FALSE
            if ("iCellR" != class(x)[1]) {
                stop("x should be an object of class iCellR")
            }
            #
            start_time1 <- Sys.time()
            #  cluster
            if(data.type == "pca") {
                DATA = (t(y@pca.data))[dims, ]
                message(paste("Clustering based on PCA data"))
            }
            #####
            message(paste("   Calculating", dist.method,"distance ..."))
            My.distances = as.matrix(dist(t(DATA),dist.method))
            #####
            ncells=dim(DATA)[2]
            cell.num = k
            #####
            message(paste("    Finding",cell.num, "neighboring cells per cell ..."))

            KNN1 = lapply(1:ncells, function(findKNN){

                MyOrd <- order(My.distances[,findKNN])[2:cell.num]
                MyDist <- My.distances[MyOrd]
                MyRoot <- rep(findKNN,cell.num-1)
                data <- cbind(MyRoot,MyOrd,MyDist)
                colnames(data)<- c("from","to","weight")
                data <- as.data.frame(data)
            })
            #####
            data <- do.call("rbind", KNN1)

            message(paste("    Generating graph from root to neighboring cells ..."))
            g <- graph.data.frame(data, directed=FALSE)
            message(paste("    Clustering ..."))
            community <- cluster_louvain(g)

            clusters = as.numeric(membership(community))
            ############ export
            clustInfo <- as.data.frame(cbind(colnames(DATA),clusters))
            row.names(clustInfo) <- clustInfo$V1
            # clustInfo <- clustInfo[-1]
            clustInfo$clusters <- as.numeric(clustInfo$clusters)
            message(paste("Found",max(unique(clustInfo$clusters)),"clusters"))
            end_time1 <- Sys.time()
            Time = difftime(end_time1,start_time1,units = "mins")
            Time = round(as.numeric(Time),digits = 2)
            message(paste("Total time",Time,"mins"))

            attributes(y)$best.clust <- clustInfo

            x@best.clust <- y@best.clust
        }

        my.obj <- x
    }
}

nDim <- ncol(my.obj@pca.data)


#######################
#######################


## PLOT IT ##
############
### PCA ###

if(doPCA){
    # calibrate the parameters so we don't have negative values for scaling later
    ncells <- nrow(cellAnnotationsFrame)
    cal.new.PCA <- data.frame(matrix(nrow=ncells))
    for (i in 1:dimens) {
        min.i <- min(my.obj@pca.data[i])
        cali.i <- (my.obj@pca.data[i] - min.i)
        cal.new.PCA[,ncol(cal.new.PCA)+1] <- cali.i
    }
    cal.new.PCA <- cal.new.PCA[,-1]
    attributes(my.obj)$pca.data <- cal.new.PCA
}

############
### TSNE ###
if (doTSNE){
    my.obj <- run.pc.tsne(my.obj, dims = 1:dimens)

    # calibrate the parameters so we don't have negative values for scaling later
    ncells <- nrow(cellAnnotationsFrame)
    cal.new.tSNE <- data.frame(matrix(nrow=ncells))
    for (j in 1:3) {
        min.i <- min(my.obj@tsne.data.3d[j])
        cali.i <- (my.obj@tsne.data.3d[j] - min.i)
        cal.new.tSNE[,ncol(cal.new.tSNE)+1] <- cali.i
    }
    cal.new.tSNE <- cal.new.tSNE[,-1]
    attributes(my.obj)$tsne.data.3d <- cal.new.tSNE
}


############
### KNETL ###

if( runKNETL ){

    # calibrate the parameters so we don't have negative values for scaling later
    ncells <- nrow(cellAnnotationsFrame)
    cal.new.UMAP <- data.frame(matrix(nrow=ncells))
    for (i in 1:3) {
        min.i <- min(my.obj@knetl.data[i]) #  umap.data[i])
        cali.i <- (my.obj@knetl.data[i] - min.i + 5) # umap.data[i] - min.i)
        cal.new.UMAP[,ncol(cal.new.UMAP)+1] <- cali.i
    }
    cal.new.UMAP <- cal.new.UMAP[,-1]
    attributes(my.obj)$knetl.data <- t(cal.new.UMAP) #  umap.data <- cal.new.UMAP
}


#################
### SAVE GAME ###
save(my.obj, file = "my.obj.Robj")
#################


##### Parameter and Feature Prep
################################

# Create the Derived Parameters Data matrices
q <- nrow(cellAnnotationsFrame)

clust <- cellAnnotationsFrame
if (isSeqGeq){
    col01 <- c("CellId")
}else{
    col01 <- c("EventNumberDP")
}

if( doClust ){
    # my.obj@best.clust[is.na(my.obj@best.clust)] <- 0
    num.clust <- max(my.obj@best.clust)
    # meta.clust <- data.frame(cluster.data[["Best.partition"]])
    write.table(num.clust, paste0("Estimated_k_", runID), row.names = F, col.names = F)
    # Data matrix for export to FJ/SG
    clust <- cbind(clust, my.obj@best.clust[1:q,])
    if (isSeqGeq){
        col01 <- c(col01,paste0("iClust_", runID))
    }else{
        col01 <- c(col01,paste0("iClust_", runID))
    }
}

if( doPCA ){
    # Data matrix for export to FJ/SG
    clust <- cbind(clust, dataPCA[1:q,1:dimens])
    for(d in 1:dimens){
        col01 <- c(col01, paste0("pc_",d,"_",runID))
    }
}

if( doTSNE ){
    # Data matrix for export to FJ/SG
    clust <- cbind(clust, my.obj@tsne.data.3d[1:q,])
    col01 <- c(col01, paste0("tSNE_X_", runID), paste0("tSNE_Y_", runID), paste0("tSNE_Z_", runID))
}

if( runKNETL ){
    # Data matrix for export to FJ/SG
    test <- my.obj@knetl.data[1:q]
    # if(doClust){
    clust <- cbind(clust, t(my.obj@knetl.data))
    # } else {
    #   clust <- t(my.obj@knetl.data)
    # }
    col01 <- c(col01, paste0("KNetL_X_", runID), paste0("KNetL_Y_", runID), paste0("KNetL_Pseudotime_", runID))
}

colnames(clust) <- col01

# Create the DP output (CSV)
FName <- paste0("/iCellR_Clust_",runID,".csv")
fileName = paste0(fjOuts,FName)

if (isSeqGeq){
    suppressWarnings(write.table(clust, file = fileName, sep = ',', append = T, quote = F, row.names = F, col.names = T ))
}else{
    FName <- paste0("/iCellR_Clust_",runID,".R")
    fileName = paste0(fjOuts,FName)
    suppressWarnings(write.table(clust, file = fileName, sep = ',', append = T, quote = F, row.names = F, col.names = T ))
}

#################################################################
#####


write.table(data.frame(1,2,3,5,8), file="fin.tsv", sep="  ", col.names=F, row.names = F)